"""
pyart.correct.vpr
=================

Computes and corrects the vertical profile of reflectivity

.. autosummary::
    :toctree: generated/

    correct_vpr
    compute_theoretical_vpr
    compute_apparent_vpr
    compute_avg
    compute_refl_time_avg
    compute_refl_ratios
    find_best_profile
    compute_vpr_correction
    filter_vpr

"""
from copy import deepcopy
from warnings import warn

import numpy as np

from ..config import get_metadata, get_field_name
from ..retrieve import get_ml_rng_limits, get_iso0_val
from ..util.radar_utils import compute_azimuthal_average
from ..util import compute_mse, cut_radar
from ..core.transforms import antenna_to_cartesian
from ..core import Radar


def correct_vpr(radar, nvalid_min=20, angle_min=0., angle_max=4.,
                ml_thickness_min=200., ml_thickness_max=800.,
                ml_thickness_step=200., iso0_max=5000.,
                ml_top_diff_max=200., ml_top_step=200., ml_peak_min=1.,
                ml_peak_max=6., ml_peak_step=1., dr_min=-6., dr_max=-1.5,
                dr_step=1.5, dr_default=-4.5, dr_alt=800., h_max=6000.,
                h_res=1., max_weight=9., rmin_obs=5000., rmax_obs=150000.,
                iso0=None, weight_mem=0.75, vpr_theo_dict_mem=None,
                radar_mem_list=None, refl_field=None, lin_refl_field=None,
                norm_refl_field=None, corr_refl_field=None, corr_field=None,
                temp_field=None, iso0_field=None, temp_ref=None):
    """
    Correct VPR using the Meteo-France operational algorithm

    Parameters
    ----------
    radar : Radar
        Radar object
    nvalid_min : int
        Minimum number of rays with data to consider the azimuthal average
        valid
    angle_min, angle_max : float
        Minimum and maximum elevation angles used to compute the ratios of
        reflectivity
    ml_thickness_min, ml_thickness_max, ml_thickness_step : float
        Minimum, maximum and step of the melting layer thickness of the models
        to explore [m]
    iso0_max : float
        maximum iso0 altitude of the profile
    ml_top_diff_max, ml_top_step : float
        maximum difference +- between iso-0 and top of the melting layer [m]
        of the models to explore. Step
    ml_peak_min, ml_peak_max, ml_peak_step: float
        min, max and step of the value at the peak of the melting layer of the
        models to explore
    dr_min, dr_max, dr_step : float
        min, max and step of the decreasing ratio above the melting layer
    dr_default : float
        default decreasing ratio to use if a proper model could not be found
    dr_alt : float
        altitude above the melting layer top where theoretical profile needs
        to be defined to be able to compute DR. If the theoretical profile is
        not defined up to the resulting altitude a default DR is used
    h_max : float
        maximum altitude [masl] where to compute the model profile
    h_res : float
        resolution of the model profile (m)
    max_weight : float
        Maximum weight of the antenna pattern
    rmin_obs, rmax_obs : float
        minimum and maximum range (m) of the observations that are compared
        with the model
    iso0 : float
        reference iso0 value
    weight_mem : float
        Weight given to the previous VPR retrieval when filtering the current
        VPR retrieval by the previous one
    vpr_theo_dict_mem : dict or None
        Dictionary containing the theoretical VPR computed in the previous
        time step
    radar_mem_list : list of radar objects or None
        list of radar objects that contain the azimuthally averaged
        reflectivity computed in the past
    refl_field : str
        Name of the reflectivity field to correct
    lin_refl_field : str
        Name of the linear reflectivity field
    norm_refl_field : str
        Name of the normalized linear reflectivity field
    corr_refl_field : str
        Name of the VPR-corrected reflectivity field
    corr_field : str
        Name of the VPR correction field
    temp_field, iso0_field : str
        Name of the temperature and height over the iso-0 fields
    temp_ref : str
        the field use as reference for temperature. Can be temperature
        or height_over_iso0.

    Returns
    -------
    refl_corr_dict : dict
        The corrected reflectivity
    corr_field_dict : dict
        The correction applied
    vpr_theo_dict_filtered : dict
        The theoretical VPR profile used for the correction
    radar_rhi : radar object
        A radar object containing the azimuthally averaged reflectivity in
        linear units

    """
    # parse the field parameters
    if refl_field is None:
        refl_field = get_field_name('reflectivity')
    if lin_refl_field is None:
        lin_refl_field = get_field_name('linear_reflectivity')
    if norm_refl_field is None:
        norm_refl_field = get_field_name('normalized_reflectivity')
    if corr_refl_field is None:
        corr_refl_field = get_field_name('corrected_reflectivity')
    if corr_field is None:
        corr_field = get_field_name('vpr_correction')
    if temp_ref == 'temperature':
        if temp_field is None:
            temp_field = get_field_name('temperature')
        temp_ref_field = temp_field
    elif temp_ref == 'height_over_iso0':
        if iso0_field is None:
            iso0_field = get_field_name('height_over_iso0')
        temp_ref_field = iso0_field

    # get azimuthally averaged data within the region of interest
    radar_azi_avg = compute_avg(
        radar, rng_min=rmin_obs, rng_max=rmax_obs, ele_min=angle_min,
        ele_max=angle_max, h_max=h_max, refl_field=refl_field,
        temp_ref_field=temp_ref_field, lin_refl_field=lin_refl_field)

    if iso0 is None:
        iso0 = get_iso0_val(
            radar_azi_avg, temp_ref_field=temp_ref_field, temp_ref=temp_ref)
    print('iso0:', iso0)

    # compute the temporal average
    radar_time_avg = compute_refl_time_avg(
        radar_azi_avg, refl_field=lin_refl_field,
        radar_mem_list=radar_mem_list, nvalid_min=nvalid_min)

    # compute ratios of reflectivity
    ele_ratios, refl_ratios = compute_refl_ratios(
        radar_time_avg, refl_field=lin_refl_field)
    print(ele_ratios)

    # find best profile
    (_, best_ml_top, best_ml_thickness, best_val_ml, best_val_dr,
     best_error) = find_best_profile(
        radar_time_avg, refl_ratios, ml_thickness_min=ml_thickness_min,
        ml_thickness_max=ml_thickness_max,
        ml_thickness_step=ml_thickness_step, iso0=iso0, iso0_max=iso0_max,
        ml_top_diff_max=ml_top_diff_max, ml_top_step=ml_top_step,
        ml_peak_min=ml_peak_min, ml_peak_max=ml_peak_max,
        ml_peak_step=ml_peak_step, dr_min=dr_min, dr_max=dr_max,
        dr_step=dr_step, dr_default=dr_default, dr_alt=dr_alt, h_max=h_max,
        h_res=h_res, max_weight=max_weight)
    print('best_ml_top', best_ml_top)
    print('best_ml_thickness', best_ml_thickness)
    print('best_val_ml', best_val_ml)
    print('best_val_dr', best_val_dr)
    print('best_error', best_error)

    # get theoretical profile as a function of altitude
    vpr_theo_dict = compute_theoretical_vpr(
        ml_top=best_ml_top, ml_thickness=best_ml_thickness,
        val_ml=best_val_ml, val_dr=best_val_dr, h_max=15000., h_res=h_res)

    # filter profile with previously found profile
    vpr_theo_dict_filtered = filter_vpr(
        vpr_theo_dict, vpr_theo_dict_mem=vpr_theo_dict_mem,
        weight_mem=weight_mem)

    # correct the reflectivity
    refl_corr_dict, corr_field_dict = compute_vpr_correction(
        radar, vpr_theo_dict_filtered, max_weight=max_weight,
        corr_field=corr_field, norm_refl_field=norm_refl_field,
        corr_refl_field=corr_refl_field, refl_field=refl_field)

    return (refl_corr_dict, corr_field_dict, vpr_theo_dict_filtered,
            radar_azi_avg)


def compute_theoretical_vpr(ml_top=3000., ml_thickness=200., val_ml=3.,
                            val_dr=-3., h_max=6000., h_res=1.):
    """
    Computes an idealized vertical profile of reflectivity

    Parameters
    ----------
    ml_top : float
        melting layer top [m asl]
    ml_thickness : float
        melting layer thickness [m]
    val_ml : float
        value in the peak of the melting layer
    val_dr : float
        decreasing ratio above the melting layer
    h_max : float
        maximum altitude at which to compute the profile [m asl]
    h_res : float
        profile resolution [m]

    Returns
    -------
    val_theo_dict : dict
        A dictionary containg the value at each altitude, the reference
        altitude and the top and bottom of the melting layer

    """
    h = np.arange(0, h_max, h_res)
    val_theo = np.ma.masked_all(h.size)

    ml_bottom = ml_top - ml_thickness
    if ml_bottom < 0.:
        ml_bottom = 0.
    ml_peak = ml_top - ml_thickness/2.

    val_theo[h < ml_bottom] = 1.
    if ml_thickness > 0:
        val_theo[(h >= ml_bottom) & (h < ml_peak)] = (
            1. + 2.*(val_ml-1.)/ml_thickness
            * (h[(h >= ml_bottom) & (h < ml_peak)]-ml_bottom))
        val_theo[(h >= ml_peak) & (h <= ml_top)] = (
            val_ml + 2.*(1-val_ml)/ml_thickness
            * (h[(h >= ml_peak) & (h <= ml_top)]-ml_peak))
        val_theo[h > ml_top] = np.exp(
            np.log(10.)/10000.*val_dr*(h[(h > ml_top)]-ml_top))
    else:
        val_theo[h >= ml_top] = np.exp(
            np.log(10.)/10000.*val_dr*(h[(h >= ml_top)]-ml_top))

    val_theo_dict = {
        'value': val_theo,
        'altitude': h,
        'ml_top': ml_top,
        'ml_bottom': ml_bottom,
        'val_ml_peak': val_ml,
        'val_dr': val_dr
    }
    return val_theo_dict


def compute_apparent_vpr(radar, ml_top=3000., ml_thickness=200., val_ml=3.,
                         val_dr=-3., h_max=6000., h_res=1., max_weight=9.,
                         vpr_theo_dict=None,
                         refl_field='normalized_reflectivity'):
    """
    Computes the apparent VPR

    Parameters
    ----------
    radar : radar object
        the reference radar object
    ml_top, ml_thickness : float
        melting layer top [m asl] and thickness [m]
    val_ml : float
        value in the peak of the melting layer
    val_dr : float
        decreasing ratio above the melting layer
    h_max : float
        maximum altitude at which to compute the theoretical profiles [m asl]
    h_res : float
        profile resolution [m]
    max_weight : float
        the factor by which the antenna beam width is multiplied
    vpr_theo_dict : dict or None
        A dictionary containing the theoretical VPR profile. If it is None the
        theoretical VPR profile is computed using the parameters describing it
    refl_field: str
        Name of the apparent reflectivity profile obtained

    Returns
    -------
    radar_out : radar object
        A radar object containing the apparent normalized reflectivity profile
    vpr_theo_dict : dict
        A dictionary containg the theoretical normalized reflectivity profile

    """
    radar_out = deepcopy(radar)
    radar_out.fields = dict()
    refl_dict = get_metadata(refl_field)
    refl_dict['data'] = np.ma.masked_all((radar_out.nrays, radar_out.ngates))
    radar_out.add_field(refl_field, refl_dict)

    rng = radar_out.range['data']
    known_ml_btm = False
    if vpr_theo_dict is None:
        # get theoretical profile as a function of altitude
        ml_bottom = ml_top - ml_thickness
        vpr_theo_dict = compute_theoretical_vpr(
            ml_top=ml_top, ml_thickness=ml_thickness, val_ml=val_ml,
            val_dr=val_dr, h_max=h_max, h_res=h_res)

        # range resolution of the radar resolution volume
        rng_res = rng[1] - rng[0]
        rng_left_km = (rng-rng_res/2.)/1000.
        rng_right_km = (rng+rng_res/2.)/1000.
        known_ml_btm = True

    beam_width = (
        radar_out.instrument_parameters['radar_beam_width_h']['data'][0])
    half_bw = beam_width/2.
    half_bw_rad = half_bw*np.pi/180.

    for ind_ray, ang in enumerate(radar_out.elevation['data']):
        if known_ml_btm:
            rng_bottom, _ = get_ml_rng_limits(
                rng_left_km, rng_right_km, rng, ang, beam_width,
                radar_out.altitude['data'][0], ml_bottom, ml_top)

            i_rng_btm = np.where(rng >= rng_bottom)[0][0]

            # values below the melting layer affected area
            radar_out.fields[refl_field]['data'][ind_ray, 0:i_rng_btm] = 1.
        else:
            i_rng_btm = 0

        rng_ml_vals = rng[i_rng_btm:] / 1000.  # km
        for i_rng, rng_val in enumerate(rng_ml_vals):
            # altitudes affected by the antenna diagram
            _, _, z_diag_min = antenna_to_cartesian(rng_val, 0., ang-half_bw)
            z_diag_min += radar_out.altitude['data']

            _, _, z_diag_max = antenna_to_cartesian(rng_val, 0., ang+half_bw)
            z_diag_max += radar_out.altitude['data']

            _, _, z_diag_center = antenna_to_cartesian(
                rng_val, 0., ang+half_bw)
            z_diag_center += radar_out.altitude['data']

            h = vpr_theo_dict['altitude'][
                (vpr_theo_dict['altitude'] >= z_diag_min)
                & (vpr_theo_dict['altitude'] <= z_diag_max)]
            vpr_vals = vpr_theo_dict['value'][
                (vpr_theo_dict['altitude'] >= z_diag_min)
                & (vpr_theo_dict['altitude'] <= z_diag_max)]

            weights = (h-z_diag_center)/(1000.*rng_val*half_bw_rad)
            weights = weights*weights
            weights = np.ma.masked_greater(weights, max_weight)
            weights = np.ma.exp(-2.*np.log(2.)*weights)

            radar_out.fields[refl_field]['data'][
                ind_ray, i_rng_btm+i_rng] = (
                    np.ma.sum(vpr_vals*weights)/np.ma.sum(weights))

    return radar_out, vpr_theo_dict


def compute_avg(radar, rng_min=5000., rng_max=150000., ele_min=0., ele_max=4.,
                h_max=6000., refl_field='reflectivity',
                temp_ref_field='height_over_iso0',
                lin_refl_field='linear_reflectivity'):
    """
    Prepares the data for VPR correction. Cuts the original radar to keep
    only the region of interest for VPR correction, transforms the
    reflectivity in linear units and makes an azimuthal average of the
    reflectivity and the reference temperature field

    Parameters
    ----------
    radar : Radar
        Radar object
    rng_min, rng_max : float
        Minimum and maximum range (m) where the data is used for VPR
        estimation
    ele_min, ele_max : float
        Minimum and maximum elevation angles where the data is used for VPR
        estimation
    h_max : float
        maximum altitude [masl] where to compute the model profile
    refl_field, tem_ref_field : str
        name of the reflectivity and temperature reference fields
    lin_refl_field : str
        name of the reflectivity in linear units field

    Returns
    -------
    radar_rhi : radar object
        a radar object with the data of interest

    """
    radar_aux = cut_radar(
        radar, [refl_field, temp_ref_field], rng_min=rng_min, rng_max=rng_max,
        ele_min=ele_min, ele_max=ele_max, azi_min=None, azi_max=None)
    if radar_aux is None:
        warn('No data within limits')
        return None

    # order sweeps by increased elevation angle
    radar_aux = radar_aux.extract_sweeps(
        np.argsort(radar_aux.fixed_angle['data']))

    # compute linear reflectivity
    refl_lin = get_metadata(lin_refl_field)
    refl_lin['data'] = np.ma.power(
        10., 0.1*radar_aux.fields[refl_field]['data'])
    radar_aux.fields.pop(refl_field)
    radar_aux.add_field(lin_refl_field, refl_lin)

    # average data in azimuth
    radar_rhi = compute_azimuthal_average(
        radar_aux, [lin_refl_field, temp_ref_field], nvalid_min=1)

    # only data below iso0_max is valid
    radar_rhi.fields[lin_refl_field]['data'][
        radar_rhi.gate_altitude['data'] >= h_max] = np.ma.masked

    return radar_rhi


def compute_refl_time_avg(radar, refl_field=None, radar_mem_list=None,
                          nvalid_min=20):
    """
    Computes the time average of the reflectivity

    Parameters
    ----------
    radar : radar object
        the radar object
    refl_field : str
        name of the reflectivity field used to compute the accumulation
    radar_mem_list : list of radar objects or None
        list of radar objects that contain the azimuthally averaged
        reflectivity computed in the past
    nvalid_min : float
        Minimum number of samples to consider the average valid

    Returns
    -------
    radar_time_avg : radar object
        A radar object containing the time-averaged reflectivity field and
        metadata

    """
    radar_arr = [radar]
    if radar_mem_list is not None:
        radar_arr = radar_arr + radar_mem_list

    # find unique elevations
    eles = np.array([])
    for radar_aux in radar_arr:
        eles = np.append(eles, np.round(radar_aux.elevation['data'], 1))
    eles = np.sort(np.unique(eles))
    nrays = eles.size

    # dummy radar
    radar_out = deepcopy(radar)
    radar_out.sweep_end_ray_index['data'] = nrays-1
    radar_out.azimuth['data'] = radar.azimuth['data'][0]+np.zeros(nrays)
    radar_out.elevation['data'] = eles

    # compute temporal average at each elevation
    refl_data = np.ma.zeros((nrays, radar.ngates))
    ns_data = np.ma.zeros((nrays, radar.ngates))
    for ind_ele, ele in enumerate(eles):
        for radar_aux in radar_arr:
            eles_aux = np.round(radar_aux.elevation['data'], 1)
            if ele not in eles_aux:
                continue
            ind_ele_aux = np.where(eles_aux == ele)[0]

            ns_data[ind_ele, :] = (
                ns_data[ind_ele, :] +
                radar_aux.fields['number_of_samples']['data'][ind_ele_aux, :])
            refl_data[ind_ele, :] = (
                refl_data[ind_ele, :]
                + radar_aux.fields[refl_field]['data'][ind_ele_aux, :]
                * ns_data[ind_ele, :])
            radar_out.time['data'][ind_ele] = radar_aux.time['data'][
                ind_ele_aux]
    refl_data /= ns_data
    refl_data[ns_data < nvalid_min] = np.ma.masked
    radar_out.fields[refl_field]['data'] = refl_data
    fields = {
        refl_field: radar_out.fields[refl_field]}

    return Radar(
        radar_out.time, radar_out.range, fields, radar_out.metadata,
        radar_out.scan_type, radar_out.latitude, radar_out.longitude,
        radar_out.altitude, radar_out.sweep_number, radar_out.sweep_mode,
        radar_out.fixed_angle, radar_out.sweep_start_ray_index,
        radar_out.sweep_end_ray_index, radar_out.azimuth, radar_out.elevation,
        instrument_parameters=radar_out.instrument_parameters)


def compute_refl_ratios(radar, refl_field=None):
    """
    Computes reflectivity ratios

    Parameters
    ----------
    radar : radar object
        the radar object
    refl_field : str
        name of the reflectivity field used to compute the ratios

    Returns
    -------
    ele_ratios : dict
        dict containing the top and bottom elevation angles of the ratios
    refl_ratios : 2D array of floats
        A 2D array which contains the ratios

    """
    ele_ratios = {
        'top': np.array([], dtype=float),
        'btm': np.array([], dtype=float),
    }
    first_ratio = True
    for ind_btm, ele_btm in enumerate(radar.elevation['data'][:-1]):
        for ele_top in radar.elevation['data'][ind_btm+1:]:
            ele_ratios['top'] = np.append(ele_ratios['top'], ele_top)
            ele_ratios['btm'] = np.append(ele_ratios['btm'], ele_btm)
            ratio_aux = np.expand_dims(
                radar.fields[refl_field]['data'][ind_btm+1, :]
                / radar.fields[refl_field]['data'][ind_btm, :], axis=0)
            if first_ratio:
                refl_ratios = ratio_aux
                first_ratio = False
            else:
                refl_ratios = np.ma.append(refl_ratios, ratio_aux, axis=0)

    return ele_ratios, refl_ratios


def find_best_profile(radar_obs, ratios_obs, ml_thickness_min=200.,
                      ml_thickness_max=800., ml_thickness_step=200.,
                      iso0=3000., iso0_max=5000., ml_top_diff_max=200.,
                      ml_top_step=200., ml_peak_min=1., ml_peak_max=6.,
                      ml_peak_step=1., dr_min=-6., dr_max=-1.5, dr_step=1.5,
                      dr_default=-4.5, dr_alt=800., h_max=6000., h_res=1.,
                      max_weight=9.,
                      norm_refl_field='normalized_reflectivity'):
    """
    gets the theoretical profile that best matches the observations for each
    elevation angle

    Parameters
    ----------
    radar_obs : radar object
        radar object containing the observed data
    ratios_obs : 2D array
        2D array field containing the ratios between higher elevation and
        lower elevations of the observed linear reflectivity
    ml_thickness_min, ml_thickness_max, ml_thickness_step : float
        Minimum, maximum and step of the melting layer thickness of the models
        to explore [m]
    iso0 : float
        iso0 [masl]
    iso0_max : float
        maximum iso0 altitude of the profile
    ml_top_diff_max, ml_top_step : float
        maximum difference +- between iso-0 and top of the melting layer [m]
        of the models to explore. Step
    ml_peak_min, ml_peak_max, ml_peak_step: float
        min, max and step of the value at the peak of the melting layer of the
        models to explore
    dr_min, dr_max, dr_step : float
        min, max and step of the decreasing ratio above the melting layer
    dr_default : float
        default decreasing ratio to set if VPR model could not be found
    dr_alt : float
        altitude above the melting layer top where theoretical profile needs
        to be defined to be able to compute DR. If the theoretical profile is
        not defined up to the resulting altitude a default DR is used
    h_max : float
        maximum altitude [masl] where to compute the model profile
    h_res : float
        resolution of the model profile (m)
    max_weight : float
        Maximum weight of the antenna pattern

    Returns
    -------
    best_vpr_theo_dict : dict
        Dictionary containing the values of the best VPR algorithm
    best_ml_top, best_ml_thickness : float
        The value of the melting layer top [masl] and thickness [m] of the
        model that closest ressembles the observations
    best_val_ml : float
        The value at the peak of the melting layer of the best model
    best_val_dr : float
        The decreasing ratio at the top of the melting layer of the best model
    best_error : the quadratic error of the best model

    """
    best_error = 1e8
    best_val_ml = -999.
    best_ml_thickness = -999.
    best_ml_top = -999.
    best_val_dr = -999.
    best_vpr_theo_dict = None

    if iso0 >= iso0_max or iso0 <= radar_obs.altitude['data']:
        warn('iso0 {} masl higher than {} m or lower than the radar altitude'.format(
            iso0, iso0_max))
        return (best_vpr_theo_dict, iso0, 0., 1., dr_default, best_error)

    ml_thickness_vals = np.arange(
        ml_thickness_min, ml_thickness_max+ml_thickness_step,
        ml_thickness_step)
    ml_top_max = iso0+ml_top_diff_max
    if ml_top_max > iso0_max:
        ml_top_max = iso0_max
    ml_top_min = iso0-ml_top_diff_max
    if ml_top_min < radar_obs.altitude['data']:
        ml_top_min = radar_obs.altitude['data']
    ml_top_vals = np.arange(
        ml_top_min, ml_top_max+ml_top_step, ml_top_step)

    ml_peak_vals = np.arange(
        ml_peak_min, ml_peak_max+ml_peak_step, ml_peak_step)
    if iso0+dr_alt >= h_max:
        warn('theoretical profile not defined high enough above the melting'
             ' layer. A default DR will be used')
        dr_vals = np.array([dr_default])
    else:
        dr_vals = np.arange(dr_min, dr_max+dr_step, dr_step)

    for ml_top in ml_top_vals:
        for ml_thickness in ml_thickness_vals:
            for val_ml in ml_peak_vals:
                for val_dr in dr_vals:
                    print('\nml top {} masl, ml thickness {}, ml peak {}, dr {}, '.format(
                              ml_top, ml_thickness, val_ml, val_dr))
                    radar_theo, vpr_theo_dict = compute_apparent_vpr(
                        radar_obs, ml_top=ml_top, ml_thickness=ml_thickness,
                        val_ml=val_ml, val_dr=val_dr, h_max=h_max, h_res=h_res,
                        max_weight=max_weight, refl_field=norm_refl_field)
                    _, ratios_theo = compute_refl_ratios(
                        radar_theo, refl_field=norm_refl_field)
                    error = compute_mse(ratios_theo, ratios_obs)
                    if error is None:
                        warn('Not enough observations to compute VPR')
                        return (best_vpr_theo_dict, iso0, 0., 1., dr_default,
                                best_error)
                    if error < best_error:
                        print('\nVALID MODEL. Error: {}\n'.format(error))
                        best_error = error
                        best_val_ml = val_ml
                        best_ml_thickness = ml_thickness
                        best_ml_top = ml_top
                        best_val_dr = val_dr
                        best_vpr_theo_dict = vpr_theo_dict

    return (best_vpr_theo_dict, best_ml_top, best_ml_thickness,
            best_val_ml, best_val_dr, best_error)


def compute_vpr_correction(radar, vpr_theo_dict, max_weight=9.,
                           corr_field='vpr_correction',
                           norm_refl_field='normalized_reflectivity',
                           corr_refl_field='corrected_reflectivity',
                           refl_field='reflectivity'):
    """
    Computes the VPR correction

    Parameters
    ----------
    radar : radar object
        reference radar object
    vpr_theo_dict : dict
        dictionary containing the theoretical VPR
    max_weight : float
        Maximum weight of the antenna pattern

    Returns
    -------
    corr_field_dict : dict
        Dictionary containing the values of the VPR correction and metadata

    """
    radar_rhi = deepcopy(radar)
    radar_rhi.fields = dict()
    radar_rhi.scan_type = 'rhi'
    radar_rhi.sweep_number['data'] = np.array([0])
    radar_rhi.sweep_mode['data'] = np.array(['rhi'])
    radar_rhi.fixed_angle['data'] = np.array([0])
    radar_rhi.sweep_start_ray_index['data'] = np.array([0])
    radar_rhi.sweep_end_ray_index['data'] = np.array([radar.nsweeps-1])
    radar_rhi.rays_per_sweep['data'] = np.array([radar.nsweeps])
    radar_rhi.azimuth['data'] = np.ones(radar.nsweeps)
    radar_rhi.elevation['data'] = radar.fixed_angle['data']
    radar_rhi.time['data'] = np.zeros(radar.nsweeps)
    radar_rhi.nrays = radar.fixed_angle['data'].size
    radar_rhi.nsweeps = 1
    radar_rhi.rays_are_indexed = None
    radar_rhi.ray_angle_res = None

    radar_corr, _ = compute_apparent_vpr(
        radar_rhi, max_weight=max_weight, vpr_theo_dict=vpr_theo_dict)

    corr_field_dict = get_metadata(corr_field)
    corr_field_dict['data'] = np.ma.zeros((radar.nrays, radar.ngates))
    for ind_ray, (ind_start, ind_end) in enumerate(zip(
                                radar.sweep_start_ray_index['data'],
                                radar.sweep_end_ray_index['data'])):
        corr_field_dict['data'][ind_start:ind_end+1, :] = 10.*np.ma.log10(
            1./radar_corr.fields[norm_refl_field]['data'][ind_ray, :])

    refl_corr_dict = get_metadata(corr_refl_field)
    refl_corr_dict['data'] = (
        radar.fields[refl_field]['data']+corr_field_dict['data'])

    return refl_corr_dict, corr_field_dict


def filter_vpr(vpr_theo_dict, vpr_theo_dict_mem=None, weight_mem=0.75):
    """
    Filters the current retrieved VPR with past retrievals

    Parameters
    ----------
    vpr_theo_dict : dict
        the current VPR retrieval
    vpr_theo_dict_mem : dict
        the past retrieval
    weight_mem : float
        weight to give to the past retrieval

    Returns
    -------
    vpr_filt_dict : dict
        The filtered retrieval
    """
    if vpr_theo_dict_mem is None:
        return deepcopy(vpr_theo_dict)
    if np.array_equal(vpr_theo_dict['altitude'],
                      vpr_theo_dict_mem['altitude']):
        vals = (
            (1.-weight_mem)*vpr_theo_dict['value']
            + weight_mem*vpr_theo_dict_mem['value'])
        vpr_filt_dict = {
            'value': vals,
            'altitude': vpr_theo_dict['altitude'],
        }
        return vpr_filt_dict
    return deepcopy(vpr_theo_dict)



"""
pyart.util.circular_stats
=========================

Functions for computing statistics on circular (directional) distributions.

.. autosummary::
    :toctree: generated/

    compute_directional_stats
    angular_mean
    angular_std
    angular_mean_deg
    angular_std_deg
    interval_mean
    interval_std
    mean_of_two_angles
    mean_of_two_angles_deg

"""

import numpy as np


# For details on these computation see:
# https://en.wikipedia.org/wiki/Directional_statistics
# https://en.wikipedia.org/wiki/Mean_of_circular_quantities


def compute_directional_stats(field, avg_type='mean', nvalid_min=1, axis=0):
    """
    Computes the mean or the median along one of the axis (ray or range)

    Parameters
    ----------
    field : ndarray
        the radar field
    avg_type :str
        the type of average: 'mean' or 'median'
    nvalid_min : int
        the minimum number of points to consider the stats valid. Default 1
    axis : int
        the axis along which to compute (0=ray, 1=range)
    
    Returns
    -------
    values : ndarray 1D
        The resultant statistics
    nvalid : ndarray 1D
        The number of valid points used in the computation
    """
    if avg_type == 'mean':
        values = np.ma.mean(field, axis=axis)
    else:
        values = np.ma.median(field, axis=axis)

    # Set to non-valid if there is not a minimum number of valid gates
    valid = np.logical_not(np.ma.getmaskarray(field))
    nvalid = np.sum(valid, axis=0, dtype=int)
    values[nvalid < nvalid_min] = np.ma.masked

    return values, nvalid


def mean_of_two_angles(angles1, angles2):
    """
    Compute the element by element mean of two sets of angles.

    Parameters
    ----------
    angles1 : array
        First set of angles in radians.
    angles2 : array
        Second set of angles in radians.

    Returns
    -------
    mean : array
        Elements by element angular mean of the two sets of angles in radians.

    """
    angles1 = np.asanyarray(angles1)
    angles2 = np.asanyarray(angles2)
    x = (np.cos(angles1) + np.cos(angles2)) / 2.
    y = (np.sin(angles1) + np.sin(angles2)) / 2.
    return np.arctan2(y, x)


def mean_of_two_angles_deg(angle1, angle2):
    """
    Compute the element by element mean of two sets of angles in degrees.

    Parameters
    ----------
    angle1 : array
        First set of angles in degrees.
    angle2 : array
        Second set of angles in degrees.

    Returns
    -------
    mean : array
        Elements by element angular mean of the two sets of angles in degrees.

    """
    return np.rad2deg(
        mean_of_two_angles(np.deg2rad(angle1), np.deg2rad(angle2)))


def angular_mean(angles):
    """
    Compute the mean of a distribution of angles in radians.

    Parameters
    ----------
    angles : array like
        Distribution of angles in radians.

    Returns
    -------
    mean : float
        The mean angle of the distribution in radians.

    """
    angles = np.asanyarray(angles)
    x = np.cos(angles)
    y = np.sin(angles)
    return np.arctan2(np.mean(y), np.mean(x))


def angular_std(angles):
    """
    Compute the standard deviation of a distribution of angles in radians.

    Parameters
    ----------
    angles : array like
        Distribution of angles in radians.

    Returns
    -------
    std : float
        Standard deviation of the distribution.

    """
    angles = np.asanyarray(angles)
    x = np.cos(angles)
    y = np.sin(angles)
    norm = np.sqrt(np.mean(x)**2 + np.mean(y)**2)
    return np.sqrt(-2 * np.log(norm))


def angular_mean_deg(angles):
    """
    Compute the mean of a distribution of angles in degrees.

    Parameters
    ----------
    angles : array like
        Distribution of angles in degrees.

    Returns
    -------
    mean : float
        The mean angle of the distribution in degrees.

    """
    return np.rad2deg(angular_mean(np.deg2rad(angles)))


def angular_std_deg(angles):
    """
    Compute the standard deviation of a distribution of angles in degrees.

    Parameters
    ----------
    angles : array like
        Distribution of angles in degrees.

    Returns
    -------
    std : float
        Standard deviation of the distribution.

    """
    return np.rad2deg(angular_std(np.deg2rad(angles)))


def interval_mean(dist, interval_min, interval_max):
    """
    Compute the mean of a distribution within an interval.

    Return the average of the array elements which are interpreted as being
    taken from a circular interval with endpoints given by interval_min and
    interval_max.

    Parameters
    ----------
    dist : array like
        Distribution of values within an interval.
    interval_min, interval_max : float
        The endpoints of the interval.

    Returns
    -------
    mean : float
        The mean value of the distribution.

    """
    # transform distribution from original interval to [-pi, pi]
    half_width = (interval_max - interval_min) / 2.
    center = interval_min + half_width
    a = (np.asarray(dist) - center) / (half_width) * np.pi

    # compute the angular mean and convert back to original interval
    a_mean = angular_mean(a)
    return (a_mean * (half_width) / np.pi) + center


def interval_std(dist, interval_min, interval_max):
    """
    Compute the standard deviation of a distribution within an interval.

    Return the standard deviation of the array elements which are interpreted
    as being taken from a circular interval with endpoints given by
    interval_min and interval_max.

    Parameters
    ----------
    dist : array_like
        Distribution of values within an interval.
    interval_min, interval_max : float
        The endpoints of the interval.

    Returns
    -------
    std : float
        The standard deviation of the distribution.

    """
    # transform distribution from original interval to [-pi, pi]
    half_width = (interval_max - interval_min) / 2.
    center = interval_min + half_width
    a = (np.asarray(dist) - center) / (half_width) * np.pi

    # compute the angular standard dev. and convert back to original interval
    a_std = angular_std(a)
    return a_std * half_width / np.pi



"""
pyart.util.sigmath
==================

Function for mathematical, signal processing and numerical routines.

.. autosummary::
    :toctree: generated/

    angular_texture_2d
    grid_texture_2d
    rolling_window
    texture
    texture_along_ray
    compute_nse
    compute_corr
    compute_mse

"""

from warnings import warn

import numpy as np
from scipy import signal, ndimage

from ..config import get_fillvalue
from .radar_utils import ma_broadcast_to


def angular_texture_2d(image, N, interval):
    """
    Compute the angular texture of an image. Uses convolutions
    in order to speed up texture calculation by a factor of ~50
    compared to using ndimage.generic_filter.

    Parameters
    ----------
    image : 2D array of floats
        The array containing the velocities in which to calculate
        texture from.
    N : int
        This is the window size for calculating texture. The texture will be
        calculated from an N by N window centered around the gate.
    interval : float
        The absolute value of the maximum velocity. In conversion to
        radial coordinates, pi will be defined to be interval
        and -pi will be -interval. It is recommended that interval be
        set to the Nyquist velocity.

    Returns
    -------
    std_dev : float array
        Texture of the radial velocity field.

    """
    # transform distribution from original interval to [-pi, pi]
    interval_max = interval
    interval_min = -interval
    half_width = (interval_max - interval_min) / 2.
    center = interval_min + half_width

    # Calculate parameters needed for angular std. dev
    im = (np.asarray(image) - center) / (half_width) * np.pi
    x = np.cos(im)
    y = np.sin(im)

    # Calculate convolution
    kernel = np.ones((N, N))
    xs = signal.convolve2d(x, kernel, mode="same", boundary="symm")
    ys = signal.convolve2d(y, kernel, mode="same", boundary="symm")
    ns = N**2

    # Calculate norm over specified window
    xmean = xs/ns
    ymean = ys/ns
    norm = np.sqrt(xmean**2 + ymean**2)
    std_dev = np.sqrt(-2 * np.log(norm)) * (half_width) / np.pi
    return std_dev


def grid_texture_2d(field, xwind=7, ywind=7):
    """
    Compute the local standard deviation of a gridded field

    Parameters
    ----------
    field : ndarray 2D
        The field over which the grid is computed in (y, x) coordinates
    xwind, ywind : int
        This is the window size in each dimension for calculating the texture.
        The texture will be calculated from a xwind by ywind window centered
        around the gate.

    Returns
    -------
    win_std : float array
        Texture of the field.

    """
    win_mean = ndimage.uniform_filter(field, (ywind, xwind))
    win_sqr_mean = ndimage.uniform_filter(field**2, (ywind, xwind))
    win_var = win_sqr_mean - win_mean**2
    ind = np.where(win_var < 0.)
    if ind[0].size > 0:
        warn('field variance contains '+str(ind[0].size) +
             ' pixels with negative variance')
    return np.sqrt(win_var)


def rolling_window(a, window):
    """ create a rolling window object for application of functions
    eg: result=np.ma.std(array, 11), 1)"""
    # create a rolling window with the data
    shape = a.shape[:-1] + (a.shape[-1] - window + 1, window)
    strides = a.strides + (a.strides[-1], )
    data_wind = np.lib.stride_tricks.as_strided(
        a, shape=shape, strides=strides)

    # create a rolling window with the mask
    mask = np.ma.getmaskarray(a)
    shape = mask.shape[:-1] + (mask.shape[-1] - window + 1, window)
    strides = mask.strides + (mask.strides[-1], )
    mask_wind = np.lib.stride_tricks.as_strided(
        mask, shape=shape, strides=strides)

    # masked rolled data
    data_wind = np.ma.masked_where(mask_wind, data_wind)

    return data_wind


def texture(radar, var):
    """ Determine a texture field using an 11pt stdev
    texarray=texture(pyradarobj, field). """
    fld = radar.fields[var]['data']
    print(fld.shape)
    tex = np.ma.zeros(fld.shape)
    for timestep in range(tex.shape[0]):
        ray = np.ma.std(rolling_window(fld[timestep, :], 11), 1)
        tex[timestep, 5:-5] = ray
        tex[timestep, 0:4] = np.ones(4) * ray[0]
        tex[timestep, -5:] = np.ones(5) * ray[-1]
    return tex


def texture_along_ray(radar, var, wind_size=7):
    """
    Compute field texture along ray using a user specified
    window size.

    Parameters
    ----------
    radar : radar object
        The radar object where the field is.
    var : str
        Name of the field which texture has to be computed.
    wind_size : int, optional
        Optional. Size of the rolling window used.

    Returns
    -------
    tex : radar field
        The texture of the specified field.

    """
    half_wind = int((wind_size-1)/2)
    fld = radar.fields[var]['data']
    tex = np.ma.zeros(fld.shape)
    tex[:] = np.ma.masked
    tex.set_fill_value(get_fillvalue())

    tex_aux = np.ma.std(rolling_window(fld, wind_size), -1)
    tex[:, half_wind:-half_wind] = tex_aux
    tex[:, 0:half_wind] = ma_broadcast_to(
        tex_aux[:, 0].reshape(tex.shape[0], 1), (tex.shape[0], half_wind))
    tex[:, -half_wind:] = ma_broadcast_to(
        tex_aux[:, -1].reshape(tex.shape[0], 1), (tex.shape[0], half_wind))

    return tex


def compute_nse(obs_data, mod_data):
    """
    Computes the Nash-Sutcliff model efficiency coefficient between observed
    and modeled profiles

    Parameters
    ----------
    obs_data, mod_data : array of floats
        The vectors on which to compute the nash

    Returns
    -------
    nash_coeff : float or None
        The Nash coefficient if it could be computed
        None otherwise

    """
    mask_obs_data = np.ma.getmaskarray(obs_data)
    mask_mod_data = np.ma.getmaskarray(mod_data)
    mask = np.logical_or(mask_obs_data, mask_mod_data)
    nvalid = np.sum(np.logical_not(mask), dtype=int)
    if nvalid == 0:
        return None
    obs_data_ma = np.ma.masked_where(mask, obs_data)
    mod_data_ma = np.ma.masked_where(mask, mod_data)

    sum_obs = np.ma.sum(obs_data_ma)
    sum_obs2 = np.ma.sum(obs_data_ma*obs_data_ma)
    diff_data = obs_data_ma-mod_data_ma
    sum_diff2 = np.ma.sum(diff_data*diff_data)

    denominator = sum_obs2 - sum_obs * sum_obs / nvalid
    if denominator <= 0.:
        return None
    return 1 - sum_diff2/denominator


def compute_corr(vec1, vec2):
    """
    Computes the Pearson correlation coefficient between two arrays

    Parameters
    ----------
    vec1, vec2 : array of floats
        The vectors on which to compute the correlation coefficient

    Returns
    -------
    corr : float or None
        The correlation coefficient

    """
    mask_vec1 = np.ma.getmaskarray(vec1)
    mask_vec2 = np.ma.getmaskarray(vec2)
    mask = np.logical_or(mask_vec1, mask_vec2)
    nvalid = np.sum(np.logical_not(mask), dtype=int)
    if nvalid == 0:
        return None
    vec1_ma = np.ma.masked_where(mask, vec1)
    vec2_ma = np.ma.masked_where(mask, vec2)

    corr = np.ma.sum(vec1_ma*vec2_ma)
    sum_vec1 = np.ma.sum(vec1_ma)
    sum_vec2 = np.ma.sum(vec2_ma)
    sum2_vec1 = np.ma.sum(vec1_ma*vec1_ma)
    sum2_vec2 = np.ma.sum(vec2_ma*vec2_ma)

    corr = (
        (corr - sum_vec1*sum_vec2/nvalid)
        / np.sqrt(sum2_vec1-sum_vec1*sum_vec1/nvalid)
        / np.sqrt(sum2_vec2-sum_vec2*sum_vec2/nvalid))
    return corr


def compute_mse(arr1, arr2):
    """
    Computes the mean square error between two arrays

    Parameters
    ----------
    vec1, vec2 : array of floats
        The vectors on which to compute the correlation coefficient

    Returns
    -------
    corr : float or None
        The correlation coefficient

    """
    mask_arr1 = np.ma.getmaskarray(arr1)
    mask_arr2 = np.ma.getmaskarray(arr2)
    mask = np.logical_or(mask_arr1, mask_arr2)
    nvalid = np.sum(np.logical_not(mask), dtype=int)
    if nvalid == 0:
        return None
    arr1_ma = np.ma.masked_where(mask, arr1)
    arr2_ma = np.ma.masked_where(mask, arr2)

    diff = arr2_ma-arr1_ma
    return np.ma.sum(diff*diff)/nvalid



"""
===================================
Linear programming phase processing
===================================

An example of using linear processing to process the differential phase
fields of a ARM C-SAPR radar.

"""
print(__doc__)

# Author: Jonathan J. Helmus (jhelmus@anl.gov)
# License: BSD 3 clause

import numpy as np
import matplotlib.pyplot as plt
import pyart

# perform LP phase processing (this takes a while)
radar = pyart.io.read_mdv('095636.mdv')

# the next line force only the first sweep to be processed, this
# significantly speeds up the calculation but should be commented out
# in production so that the entire volume is processed
radar = radar.extract_sweeps([0])

phidp, kdp = pyart.correct.phase_proc_lp(radar, 0.0, debug=True)
radar.add_field('corrected_differential_phase', phidp)
radar.add_field('corrected_specific_diff_phase', kdp)

# the following line can be used to save/load in preprocessed data
#pyart.io.write_cfradial('preprocessed.nc', radar)
#radar = pyart.io.read_cfradial('preprocessed.nc')

# create a plot of the various differential phase fields
display = pyart.graph.RadarDisplay(radar)
fig = plt.figure(figsize=(10, 10))
ax1 = fig.add_subplot(221)
display.plot('differential_phase', 0, ax=ax1,
             title='Raw Differential Phase', colorbar_label='',
             axislabels_flag=False)

ax2 = fig.add_subplot(222)
display.plot('corrected_differential_phase', 0, ax=ax2,
             title='Processed Differential Phase', colorbar_label='',
             axislabels_flag=False)

ax3 = fig.add_subplot(223)
display.plot('specific_differential_phase', 0, ax=ax3, vmin=0, vmax=6,
             title='Raw Specific Differential Phase', colorbar_label='',
             axislabels_flag=False)

ax4 = fig.add_subplot(224)
display.plot('corrected_specific_diff_phase', 0, ax=ax4, vmin=0, vmax=6,
             title='Processed Specific Differential Phase',
             colorbar_label='',
             axislabels_flag=False)
plt.show()

# plot a fields from a single ray
display = pyart.graph.RadarDisplay(radar)
fig = plt.figure(figsize=[10, 5])
ax = fig.add_subplot(111)
ray_num = 191

# filtered phidp and unfolded phidp
display.plot_ray('corrected_differential_phase', ray_num, format_str='b-',
                 axislabels_flag=False, title_flag=False, ax=ax)

display.plot_ray('unfolded_differential_phase', ray_num, format_str='g-',
                 axislabels_flag=False, title_flag=False, ax=ax)

# set labels
ax.set_ylim(0, 250)
ax.set_ylabel('Differential phase shift (degrees)')
ax.set_xlabel('Range (km)')

# plot KDP and reflectivity on second axis
ax2 = ax.twinx()
display.plot_ray('corrected_specific_diff_phase', ray_num, format_str='r-',
                 axislabels_flag=False, title_flag=False, ax=ax2)
radar.add_field_like('reflectivity', 'scaled_reflectivity',
                     radar.fields['reflectivity']['data']/10.)
display.plot_ray('scaled_reflectivity', ray_num, format_str='b-',
                 axislabels_flag=False, title_flag=False, ax=ax2)

# decorate
ax2.yaxis.grid(color='gray', linestyle='dashed')
ax.legend(display.plots,
          ["Filtered phiDP", "Unfolded phiDP", 'KDP', 'Z/10.0'],
          loc='upper left')
plt.show()



"""
pyart.aux_io.mf_png_reader
==========================

Routines for putting MeteoFrance operational radar data contained in png files
into grid object.

.. autosummary::
    :toctree: generated/

    read_png
    _get_datatype_from_file
    _get_physical_data

"""

import os
import datetime
from warnings import warn
from copy import deepcopy

import numpy as np

# check existence of imageio
try:
    from imageio import imread
    _IMAGEIO_AVAILABLE = True
except ImportError:
    _IMAGEIO_AVAILABLE = False

from ..config import FileMetadata
from ..io.common import _test_arguments
from ..core.grid import Grid
from ..exceptions import MissingOptionalDependency
from ..util import ma_broadcast_to

PNG_FIELD_NAMES = {
    'hail': 'precipitation_type',
    'snow': 'precipitation_type',
    'rain': 'precipitation_type',
    'ice': 'precipitation_type'
}


def read_png(filename, additional_metadata=None, xres=1.25, yres=1.25, nx=1840,
             ny=1670, nz=1, **kwargs):

    """
    Read a MeteoFrance data png file.

    Parameters
    ----------
    filename : str
        Name of the file to read.
    additional_metadata : dict of dicts, optional
        Dictionary of dictionaries to retrieve metadata during this read.
        This metadata is not used during any successive file reads unless
        explicitly included.  A value of None, the default, will not
        introduct any addition metadata and the file specific or default
        metadata as specified by the Py-ART configuration file will be used.
    xres, yres : float
        resolution of each grid point [km]
    nx, ny, nz : int
        dimensions of the grid

    Returns
    -------
    grid : Grid
        Grid object containing the data.

    """
    # check that wradlib is available
    if not _IMAGEIO_AVAILABLE:
        raise MissingOptionalDependency(
            "imageio is required to use read_png but is not installed")

    # test for non empty kwargs
    _test_arguments(kwargs)

    try:
        ret = imread(filename, format='png', pilmode='I')
    except (OSError, SyntaxError):
        warn('Unable to read file '+filename)
        return None

    # reserved_variables = [
    #     'time', 'x', 'y', 'z',
    #     'origin_latitude', 'origin_longitude', 'origin_altitude',
    #     'point_x', 'point_y', 'point_z', 'projection',
    #     'point_latitude', 'point_longitude', 'point_altitude',
    #     'radar_latitude', 'radar_longitude', 'radar_altitude',
    #     'radar_name', 'radar_time', 'base_time', 'time_offset',
    #     'ProjectionCoordinateSystem']

    # metadata
    filemetadata = FileMetadata('PNG', PNG_FIELD_NAMES, additional_metadata)

    # required reserved variables
    time = filemetadata('grid_time')
    origin_latitude = filemetadata('origin_latitude')
    origin_longitude = filemetadata('origin_longitude')
    origin_altitude = filemetadata('origin_altitude')
    x = filemetadata('x')
    y = filemetadata('y')
    z = filemetadata('z')

    x['data'] = 1000.*(np.arange(nx)*xres+xres/2.)-889375.418
    y['data'] = 1000.*(np.arange(ny)*yres+yres/2.)+4794775.243
    z['data'] = np.array([0.])

    # origin of webmercator
    origin_latitude['data'] = np.array([0.])
    origin_longitude['data'] = np.array([0.])
    origin_altitude['data'] = np.array([0.])

    # projection (web mercator)
    projection = {
        'proj': 'webmerc',
        'ellps': 'WGS84',
        'datum': 'WGS84',
    }

    bfile = os.path.basename(filename).split('.')[0]
    # Time
    prod_time = datetime.datetime.strptime(bfile.split('_')[-1], '%Y%m%d')
    time['units'] = 'seconds since '+prod_time.strftime('%Y-%m-%d 00:00:00')
    time['data'] = np.array([0])

    # read in the fields
    datatype = _get_datatype_from_file(bfile)
    if datatype is None:
        return None

    fields = {}
    field = filemetadata.get_field_name(datatype)
    field_dict = filemetadata(field)
    field_dict['data'] = ma_broadcast_to(
        _get_physical_data(ret, datatype)[::-1, :], (nz, ny, nx))
    fields[field] = field_dict

    # radar variables
    radar_latitude = None
    radar_longitude = None
    radar_altitude = None
    radar_name = None
    radar_time = None

    return Grid(
        time, fields, dict(), origin_latitude, origin_longitude,
        origin_altitude, x, y, z, projection=projection,
        radar_latitude=radar_latitude, radar_longitude=radar_longitude,
        radar_altitude=radar_altitude, radar_name=radar_name,
        radar_time=radar_time)


def _get_datatype_from_file(filename):
    """
    gets data type from file name

    Parameters
    ----------
    filename : str
        base name of the file

    Returns
    -------
    datatype : str
        Data type contained in the file

    """
    if 'GRELE' in filename:
        datatype = 'hail'
    elif 'PLUIE' in filename:
        datatype = 'rain'
    elif 'NEIGE' in filename:
        datatype = 'snow'
    elif 'VERGLAS' in filename:
        datatype = 'ice'
    else:
        warn('Unknown HYDRE product')
        return None
    return datatype


def _get_physical_data(data, datatype):
    """
    gets data in physical units

    Parameters
    ----------
    data : int ndarray
        the data as int
    datatype : str
        The data type

    Returns
    -------
    data_ph : float ndarray
        the data in physical units

    """
    data_ph = deepcopy(data)

    #   0 Précipitations indétectables/Absence de neige au sol
    #   1 Précipitations indétectables/Présence de neige au sol
    #   2 Précipitations indétectables/Sol invisible (nuit ou nuages)
    #   3 Bruine
    #   4 Bruine sur sol gelé
    #   5 Bruine verglaçante
    #   6 Pluie
    #   7 Pluie sur sol gelé
    #   8 Pluie verglaçante
    #   9 Pluie et neige mêlées
    #  10 Neige mouillée
    #  11 Neige mouillée tenant au sol
    #  12 Neige humide
    #  13 Neige humide tenant au sol
    #  14 Neige sèche
    #  15 Neige sèche tenant au sol
    #  16 Granules de glace
    #  17 Grésil/Petite grêle
    #  18 Moyenne grêle
    #  19 Grosse grêle
    # 255 Indéterminé

    # Precip not detected
    data_ph[data == 0] = 0

    if datatype == 'rain':
        # rain categories
        data_ph[data == 174] = 3
        data_ph[data == 113] = 6
    elif datatype == 'ice':
        # ice categories
        data_ph[data == 77] = 4
        data_ph[data == 166] = 5

        data_ph[data == 121] = 7
        data_ph[data == 216] = 8
    elif datatype == 'snow':
        # # snow categories
        data_ph[data == 105] = 9  # rain/snow

        data_ph[data == 15] = 10  # WS
        data_ph[data == 101] = 11  # WS on ground

        data_ph[data == 26] = 12  # HS
        data_ph[data == 143] = 13  # HS on ground

        data_ph[data == 113] = 14  # DS
        data_ph[data == 191] = 15  # DS on ground
    elif datatype == 'hail':
        # hail categories
        data_ph[data == 226] = 17
        data_ph[data == 189] = 18
        data_ph[data == 151] = 19

    # no data
    data_ph[data == 171] = 20

    return data_ph



"""
pyart.util.datetime_utils
=========================

Functions for converting date and time between various forms.

.. autosummary::
    :toctree: generated/
    datetime_from_radar
    datetimes_from_radar
    datetime_from_dataset
    datetimes_from_dataset
    datetime_from_grid

"""

try:
    from cftime import num2date, date2num
except ImportError:
    from netCDF4 import num2date, date2num

EPOCH_UNITS = "seconds since 1970-01-01T00:00:00Z"


def datetime_from_radar(radar, epoch=False, **kwargs):
    """ Return a datetime for the first ray in a Radar. """
    if epoch:
        dtrad = num2date(radar.time['data'][0], radar.time['units'])
        epnum = date2num(dtrad, EPOCH_UNITS)
        return num2date(epnum, EPOCH_UNITS, **kwargs)
    else:
        return num2date(radar.time['data'][0], radar.time['units'],
                        **kwargs)


def datetimes_from_radar(radar, epoch=False, **kwargs):
    """ Return an array of datetimes for the rays in a Radar. """
    if epoch:
        dtrad = num2date(radar.time['data'][:], radar.time['units'])
        epnum = date2num(dtrad, EPOCH_UNITS)
        return num2date(epnum, EPOCH_UNITS, **kwargs)
    else:
        return num2date(radar.time['data'][:], radar.time['units'],
                        **kwargs)


def datetime_from_dataset(dataset, epoch=False, **kwargs):
    """ Return a datetime for the first time in a netCDF Dataset. """
    if epoch:
        dtdata = num2date(dataset.variables['time'][0],
                          dataset.variables['time'].units)
        epnum = date2num(dtdata, EPOCH_UNITS)
        return num2date(epnum, EPOCH_UNITS, **kwargs)
    else:
        return num2date(dataset.variables['time'][0],
                        dataset.variables['time'].units, **kwargs)


def datetimes_from_dataset(dataset, epoch=False, **kwargs):
    """ Return an array of datetimes for the times in a netCDF Dataset. """
    if epoch:
        dtdata = num2date(dataset.variables['time'][:],
                          dataset.variables['time'].units)
        epnum = date2num(dtdata, EPOCH_UNITS)
        return num2date(epnum, EPOCH_UNITS, **kwargs)
    else:
        return num2date(dataset.variables['time'][:],
                        dataset.variables['time'].units, **kwargs)


def datetime_from_grid(grid, epoch=False, **kwargs):
    """ Return a datetime for the volume start in a Grid. """
    if epoch:
        dtrad = num2date(grid.time['data'][0], grid.time['units'])
        epnum = date2num(dtrad, EPOCH_UNITS)
        return num2date(epnum, EPOCH_UNITS, **kwargs)
    else:
        return num2date(grid.time['data'][0], grid.time['units'],
                        **kwargs)



""" Unit Tests for Py-ART's correct/unwrap.py module. """

# python test_dealias.py
# to recreate a dealias_plot.png file showing the before and after doppler
# velocities

import numpy as np
from numpy.testing import assert_allclose, assert_almost_equal
import pytest

import pyart


REF_DATA = [0.5, 1.5, 2.5, 3.5, 4.5, 5.5, 6.5, 7.5, 8.5, 9.5, 10.5, 11.5,
            12.5, 13.5, 12.5, 11.5, 10.5, 9.5, 8.5, 7.5, 6.5, 5.5, 4.5, 3.5,
            2.5, 1.5, 0.5]


def test_dealias_unwrap_phase_ray():
    radar, dealias_vel = perform_dealias('ray')
    assert_allclose(dealias_vel['data'][13, :27], REF_DATA)
    assert np.ma.is_masked(dealias_vel['data'][13]) is False


def test_dealias_unwrap_phase_sweep():
    radar, dealias_vel = perform_dealias('sweep')
    assert_allclose(dealias_vel['data'][13, :27], REF_DATA)
    assert np.ma.is_masked(dealias_vel['data'][13]) is False


def test_dealias_unwrap_phase_volume():
    radar, dealias_vel = perform_dealias('volume')
    assert_allclose(dealias_vel['data'][13, :27], REF_DATA)
    assert np.ma.is_masked(dealias_vel['data'][13]) is False


def test_set_limits():

    radar, dealias_vel = perform_dealias('ray', set_limits=True)
    assert 'valid_min' in dealias_vel
    assert_almost_equal(dealias_vel['valid_min'], -30.0)
    assert 'valid_max' in dealias_vel
    assert_almost_equal(dealias_vel['valid_max'], 30.0)

    radar, dealias_vel = perform_dealias('ray', set_limits=False)
    assert 'valid_min' not in dealias_vel
    assert 'valid_max' not in dealias_vel


def test_dealias_unwrap_phase_no_gatefilter():
    radar, dealias_vel = perform_dealias('sweep', gatefilter=False)
    assert_allclose(dealias_vel['data'][13, :27], REF_DATA)
    assert np.ma.is_masked(dealias_vel['data'][13]) is False


def test_dealias_unwrap_phase_explicit_gatefilter():
    radar = pyart.testing.make_velocity_aliased_radar()
    gf = pyart.correct.GateFilter(radar)
    radar.fields['velocity']['data'][13, -4:] = [-7.5, 8.5, 0, 0]
    dealias_vel = pyart.correct.dealias_unwrap_phase(radar, gatefilter=gf)
    assert_allclose(dealias_vel['data'][13, :27], REF_DATA)
    assert np.ma.is_masked(dealias_vel['data'][13]) is False


def test_dealias_unwrap_phase_masked_field_sweep():
    radar = pyart.testing.make_velocity_aliased_radar()
    vdata = radar.fields['velocity']['data']
    radar.fields['velocity']['data'] = np.ma.masked_array(vdata)
    radar.fields['velocity']['data'][13, -4:] = [-7.5, 8.5, 0, 0]
    radar.fields['velocity']['data'][180, 25] = np.ma.masked
    dealias_vel = pyart.correct.dealias_unwrap_phase(
        radar, unwrap_unit='volume')
    assert_allclose(dealias_vel['data'][13, :27], REF_DATA)
    assert np.ma.is_masked(dealias_vel['data'][13]) is False
    assert np.ma.is_masked(dealias_vel['data'][180, 25])
    assert not np.ma.is_masked(dealias_vel['data'][180, 24])


def test_dealias_unwrap_phase_masked_field_volume():
    radar = pyart.testing.make_velocity_aliased_radar()
    vdata = radar.fields['velocity']['data']
    radar.fields['velocity']['data'] = np.ma.masked_array(vdata)
    radar.fields['velocity']['data'][13, -4:] = [-7.5, 8.5, 0, 0]
    radar.fields['velocity']['data'][180, 25] = np.ma.masked
    dealias_vel = pyart.correct.dealias_unwrap_phase(
        radar, unwrap_unit='volume')
    assert_allclose(dealias_vel['data'][13, :27], REF_DATA)
    assert np.ma.is_masked(dealias_vel['data'][13]) is False
    assert np.ma.is_masked(dealias_vel['data'][180, 25])
    assert not np.ma.is_masked(dealias_vel['data'][180, 24])


def test_dealias_unwrap_phase_masked_field_ray():
    radar = pyart.testing.make_velocity_aliased_radar()
    vdata = radar.fields['velocity']['data']
    radar.fields['velocity']['data'] = np.ma.masked_array(vdata)
    radar.fields['velocity']['data'][13, -4:] = [-7.5, 8.5, 0, 0]
    radar.fields['velocity']['data'][180, 25] = np.ma.masked
    dealias_vel = pyart.correct.dealias_unwrap_phase(
        radar, unwrap_unit='ray')
    assert_allclose(dealias_vel['data'][13, :27], REF_DATA)
    assert np.ma.is_masked(dealias_vel['data'][13]) is False
    assert np.ma.is_masked(dealias_vel['data'][180, 25])
    assert not np.ma.is_masked(dealias_vel['data'][180, 24])


def test_dealias_unwrap_phase_rhi_sweep():
    radar = pyart.testing.make_velocity_aliased_rhi_radar()
    radar.fields['velocity']['data'][13, -4:] = [-7.5, 8.5, 0, 0]
    dealias_vel = pyart.correct.dealias_unwrap_phase(radar)
    assert_allclose(dealias_vel['data'][13, :27], REF_DATA)
    assert np.ma.is_masked(dealias_vel['data'][13]) is False


def test_dealias_unwrap_phase_rhi_volume():
    radar = pyart.testing.make_velocity_aliased_rhi_radar()
    radar.fields['velocity']['data'][13, -4:] = [-7.5, 8.5, 0, 0]
    dealias_vel = pyart.correct.dealias_unwrap_phase(
        radar, unwrap_unit='volume')
    assert_allclose(dealias_vel['data'][13, :27], REF_DATA)
    assert np.ma.is_masked(dealias_vel['data'][13]) is False


def test_dealias_keep_original():
    radar = pyart.testing.make_velocity_aliased_radar()
    radar.fields['velocity']['data'][180, 5] = 88
    gf = pyart.filters.GateFilter(radar)
    gf.exclude_above('velocity', 40)

    dealias_vel = pyart.correct.dealias_unwrap_phase(
        radar, gatefilter=gf, keep_original=False)
    assert np.ma.is_masked(dealias_vel['data'][180, 5]) is True

    dealias_vel = pyart.correct.dealias_unwrap_phase(
        radar, gatefilter=gf, keep_original=True)
    assert_almost_equal(dealias_vel['data'][180, 5], 88)
    assert np.ma.is_masked(dealias_vel['data'][180, 5]) is False


def test_is_radar_sweep_aligned():
    radar = pyart.testing.make_empty_ppi_radar(3, 4, 2)
    radar.scan_type = 'rhi'
    radar.elevation['data'][:] = [1, 2, 3, 4, 1, 2, 3, 4]
    assert pyart.correct.unwrap._is_radar_sweep_aligned(radar)

    radar.elevation['data'][:] = [1, 2, 3, 4, 2, 3, 4, 5]
    assert not pyart.correct.unwrap._is_radar_sweep_aligned(radar)

    # raises ValueError
    radar = pyart.testing.make_empty_ppi_radar(3, 4, 2)
    radar.scan_type = 'fuzz'
    pytest.raises(
        ValueError, pyart.correct.unwrap._is_radar_sweep_aligned, radar)


def test_is_sweep_sequential():
    radar = pyart.testing.make_empty_ppi_radar(3, 4, 2)
    radar.scan_type = 'vpt'
    assert pyart.correct.unwrap._is_sweep_sequential(radar, 0)


def test_dealias_unwrap_phase_raises():

    # invalid unwrap_unit
    radar = pyart.testing.make_velocity_aliased_radar()
    pytest.raises(ValueError, pyart.correct.dealias_unwrap_phase, radar,
                  unwrap_unit='fuzz')

    # no explicit nyquist
    radar = pyart.testing.make_velocity_aliased_radar()
    radar.instrument_parameters = None
    pytest.raises(LookupError, pyart.correct.dealias_unwrap_phase, radar)

    # non-sequential
    radar = pyart.testing.make_velocity_aliased_radar()
    radar.azimuth['data'][10] = 190.
    pytest.raises(ValueError, pyart.correct.dealias_unwrap_phase, radar)

    # non-aligned sweeps
    radar = pyart.testing.make_empty_ppi_radar(1, 10, 2)
    radar.fields['velocity'] = {'data': np.zeros((20, 1))}
    pytest.raises(ValueError, pyart.correct.dealias_unwrap_phase, radar,
                  nyquist_vel=10, unwrap_unit='volume')

    # non-cubic radar
    radar = pyart.testing.make_empty_ppi_radar(1, 10, 2)
    radar.fields['velocity'] = {'data': np.zeros((20, 1))}
    radar.azimuth['data'][-10:] = range(10)
    radar.sweep_end_ray_index['data'][-1] = 18
    pytest.raises(ValueError, pyart.correct.dealias_unwrap_phase, radar,
                  nyquist_vel=10, unwrap_unit='volume')

    # invalid scan type
    radar = pyart.testing.make_velocity_aliased_radar()
    radar.scan_type = 'fuzz'
    pytest.raises(ValueError, pyart.correct.dealias_unwrap_phase, radar)

    # invalid scan type
    radar = pyart.testing.make_velocity_aliased_radar()
    radar.scan_type = 'fuzz'
    pytest.raises(ValueError, pyart.correct.dealias_unwrap_phase, radar)


def perform_dealias(unwrap_unit='sweep', **kwargs):
    """ Perform velocity dealiasing on reference data. """
    radar = pyart.testing.make_velocity_aliased_radar()
    # speckling that will not be not be dealiased.
    radar.fields['velocity']['data'][13, -4:] = [-7.5, 8.5, 0, 0]
    dealias_vel = pyart.correct.dealias_unwrap_phase(
        radar, unwrap_unit=unwrap_unit, **kwargs)
    return radar, dealias_vel


def test_fillvalue(**kwargs):
    radar = pyart.testing.make_velocity_aliased_radar()
    radar.fields['velocity']['_FillValue'] = 8888.0

    # ensure that at least one gate is excluded and should be masked
    radar.fields['velocity']['data'][100, 25] = 44
    gf = pyart.filters.GateFilter(radar)
    gf.exclude_above('velocity', 40)

    dealias_vel = pyart.correct.dealias_unwrap_phase(radar, gatefilter=gf)
    assert isinstance(dealias_vel['data'], np.ma.MaskedArray)
    assert np.ma.is_masked(dealias_vel['data'][100, 25])

    assert '_FillValue' in dealias_vel
    assert dealias_vel['_FillValue'] == 8888.0
    assert dealias_vel['data'].fill_value == 8888.0


if __name__ == "__main__":

    radar, dealias_vel = perform_dealias()
    radar.fields['dealiased_velocity'] = dealias_vel

    # print out results
    print("ray 13 velocitites before dealias:")
    print(radar.fields['velocity']['data'][13])
    print("ray 13 velocities after dealias:")
    print(radar.fields['dealiased_velocity']['data'][13])

    # create plot
    import matplotlib.pyplot as plt
    fig = plt.figure(figsize=[5, 10])
    ax1 = fig.add_subplot(2, 1, 1)
    ax2 = fig.add_subplot(2, 1, 2)
    rd = pyart.graph.RadarDisplay(radar)
    rd.plot_ppi('velocity', 0, ax=ax1, colorbar_flag=False,
                title='', vmin=-10, vmax=10)
    rd.plot_ppi('dealiased_velocity', 0, ax=ax2, colorbar_flag=False,
                title='', vmin=-10, vmax=20)
    fig.savefig('dealias_plot.png')



"""
pyart.util.ivic
===============

Estimation of noise in a ray using Ivic (2013) method

.. autosummary::
    :toctree: generated/

    estimate_noise_ivic13
    get_ivic_pct
    _func_pct_int
    _func_pct
    get_ivic_flat_reg_var_max
    _func_flat_reg
    get_ivic_snr_thr
    ivic_pct_table
    ivic_flat_reg_var_max_table
    ivic_flat_reg_wind_len_table
    ivic_snr_thr_table


"""

from copy import deepcopy
import numpy as np
from scipy.special import gammainc, gammaincc, gammaln, gammainccinv
from scipy import integrate
from scipy.optimize import fsolve

from .sigmath import rolling_window
from .radar_utils import ma_broadcast_to

def estimate_noise_ivic13(pwr_w_ray, pct=3.270436, delay=2, flat_reg_wlen=32,
                          flat_reg_var_max=0.439551, snr_thr=1.769572,
                          npulses=30, ngates_min=800, ngates_final_min=200,
                          ngates_median=10,
                          iterations=10,
                          get_noise_pos=False):
    """
    Estimate noise parameters of a ray

    Use the method of estimating the noise level in a ray outlined
    by Ivic, 2013.

    Parameters
    ----------
    pwr_w_ray : array like
        Doppler spectrum in linear units.
    pct : float
        Point Clutter Threshold
    delay : int
        distance of the gate to which compare for point target threshold
    flat_reg_wlen : int
        Minimum number of gates that should contain a valid region. Default
        gives a size of 8 km with 83.3 m resolution
    flat_reg_var_max : float
        Maximum local variance of powers in decibels to consider the region
        as flat.
    snr_thr : float
        Threshold applied in steps 3 and 6
    npulses : int
        Number of pulses used to compute the power of the array
    ngates_min: int
        minimum number of gates with noise to consider the retrieval valid
    ngates_final_min : int
        minimum number of gates that have to be left after the last step to
        consider the retrieval valid
    ngates_median : int
        number of consecutive gates above the median power that would result
        in them considered signal and removed
    iterations: int
        number of iterations in step 7
    get_noise_pos : bool
        If True the indices of the gates identified as noise will be returned

    Returns
    -------
    mean : float-like
        Mean of the gates in the ray identified as noise.
    var : float-like
        Variance of the gates in the ray identified as noise.
    nnoise : int
        Number of noise gates in the ray.
    inds_ray : 1D-array
        The indices of the gates containing noise

    References
    ----------
    I.R. Ivic, C. Curtis and S.M. Torres, Radial-Based Noise Power Estimation
    for Weather Radars. Journal of Atmospheric and Oceanic Technology, 2013,
    30, 2737-2753.

    """
    half_flat_reg_wlen = int((flat_reg_wlen-1)/2)

    if get_noise_pos:
        inds_ray = np.ma.arange(pwr_w_ray.size)

    # step 1 remove gates with discontinuities in range
    pwr_w_ray_aux = pwr_w_ray[delay:-delay]
    pwr_w_ray_aux[np.logical_or(
        np.ma.greater(pwr_w_ray_aux, pct*pwr_w_ray[:-2*delay]),
        np.ma.greater(pwr_w_ray_aux, pct*pwr_w_ray[2*delay:]))] = np.ma.masked
    if get_noise_pos:
        inds_ray_aux = inds_ray[delay:-delay]
        inds_ray_aux[np.ma.getmaskarray(pwr_w_ray_aux)] = np.ma.masked
        inds_ray = np.ma.array(inds_ray_aux.compressed())
    pwr_w_ray = np.ma.array(pwr_w_ray_aux.compressed())

    if pwr_w_ray.size*npulses < ngates_min:
        return None, None, None, None

    # step 2 detect flat sections and estimate mean power of the grouped flat
    # sections. Take smallest power as intermediate power

    # compute moving variance (in dB)
    pwr_dB_ray = np.ma.log10(pwr_w_ray)
    pwr_dB_wind = rolling_window(pwr_dB_ray, flat_reg_wlen)
    pwr_dB_wind_mean = np.ma.mean(pwr_dB_wind, axis=-1)
    # add third dimension made of mean values of the window
    pwr_dB_wind_mean = ma_broadcast_to(
        np.expand_dims(pwr_dB_wind_mean, axis=-1),
        (pwr_dB_wind_mean.size, flat_reg_wlen))
    pwr_dB_var = np.ma.sum(
        (pwr_dB_wind-pwr_dB_wind_mean)*(pwr_dB_wind-pwr_dB_wind_mean),
        axis=-1)

    # Compute intermediate power based on flat areas
    # mask non-flat regions
    pwr_w_ray_aux = deepcopy(
        pwr_w_ray[half_flat_reg_wlen:-half_flat_reg_wlen])
    inds_non_flat = np.ma.where(pwr_dB_var > flat_reg_var_max)[0]
    for ind in inds_non_flat:
        pwr_w_ray_aux[ind-half_flat_reg_wlen:ind+half_flat_reg_wlen+1] = (
            np.ma.masked)

    # group consecutive gates and get the minimum mean power
    mask = np.ma.getmaskarray(pwr_w_ray_aux)
    ind = np.ma.where(mask == False)[0]
    cons_list = np.split(ind, np.where(np.diff(ind) != 1)[0]+1)
    pwr_w_int = None
    for cons in cons_list:
        pwr_w_mean = np.ma.mean(pwr_w_ray_aux[cons])
        if pwr_w_int is None:
            pwr_w_int = pwr_w_mean
        elif pwr_w_int > pwr_w_mean:
            pwr_w_int = pwr_w_mean

    if pwr_w_int is None:
        return None, None, None, None

    # step 3 remove gates exceeding a threshold based on pwr_int
    ind = np.ma.where(pwr_w_ray > pwr_w_int*snr_thr)[0]
    if ind.size > 0:
        pwr_w_ray[ind] = np.ma.masked
    if get_noise_pos:
        inds_ray[np.ma.getmaskarray(pwr_w_ray)] = np.ma.masked
        inds_ray = np.ma.array(inds_ray.compressed())
    pwr_w_ray = np.ma.array(pwr_w_ray.compressed())
    if pwr_w_ray.size*npulses < ngates_min:
        return None, None, None, None

    # step 4 & 5: detect ngates_median or more consecutive gates with power
    # larger than the median. Compute the mean of the remaining data
    ind = np.ma.where(pwr_w_ray > np.ma.median(pwr_w_ray))[0]
    cons_list = np.split(ind, np.where(np.diff(ind) != 1)[0]+1)
    for cons in cons_list:
        if len(cons) > ngates_median:
            pwr_w_ray[cons] = np.ma.masked
    if get_noise_pos:
        inds_ray[np.ma.getmaskarray(pwr_w_ray)] = np.ma.masked
        inds_ray = np.ma.array(inds_ray.compressed())
    pwr_w_ray = np.ma.array(pwr_w_ray.compressed())
    if pwr_w_ray.size*npulses < ngates_min:
        return None, None, None, None

    pwr_w_mean = np.ma.mean(pwr_w_ray)

    # step 6: same as test 3 but with pwr_w_mean
    ind = np.ma.where(pwr_w_ray > pwr_w_mean*snr_thr)[0]
    if ind.size > 0:
        pwr_w_ray[ind] = np.ma.masked
    if get_noise_pos:
        inds_ray[np.ma.getmaskarray(pwr_w_ray)] = np.ma.masked
        inds_ray = np.ma.array(inds_ray.compressed())
    pwr_w_ray = np.ma.array(pwr_w_ray.compressed())
    if pwr_w_ray.size*npulses < ngates_min:
        return None, None, None, None

    # step 7: running sum
    rs_wlen = int(round(500/npulses))
    if rs_wlen % 2 == 0:
        rs_wlen += 1
    half_rs_wlen = int((rs_wlen-1)/2)
    criterion = gammaincc(rs_wlen*npulses, 1.12*rs_wlen*npulses)

    for _ in range(iterations):
        # check if there are sufficient noisy gates left
        if pwr_w_ray.compressed().size*npulses < ngates_final_min:
            # exit with no estimate if after first iteration there are less
            # than ngates_final_min samples left
            return None, None, None, None

        # compute running sum
        pwr_mw_wind = rolling_window(pwr_w_ray, rs_wlen)
        rs = np.ma.sum(pwr_mw_wind, axis=-1)
        pwr_w_ray = pwr_w_ray[half_rs_wlen:-half_rs_wlen]

        # check if sums exceeding threshold are below the criteria
        inds_excess = np.ma.where(rs > 1.12*rs_wlen*np.ma.mean(pwr_w_ray))[0]
        if inds_excess.size/pwr_w_ray.size <= criterion:
            # The ratio of invalid running sums is below the criterion
            break

        for ind in inds_excess:
            pwr_w_ray[ind-half_rs_wlen:ind+half_rs_wlen+1] = np.ma.masked
        if get_noise_pos:
            inds_ray = inds_ray[half_rs_wlen:-half_rs_wlen]
            inds_ray[np.ma.getmaskarray(pwr_w_ray)] = np.ma.masked
            inds_ray = np.ma.array(inds_ray.compressed())
        pwr_w_ray = np.ma.array(pwr_w_ray.compressed())

    if get_noise_pos:
        inds_ray = inds_ray.compressed()
    else:
        inds_ray = None

    return (
        np.ma.mean(pwr_w_ray), np.ma.var(pwr_w_ray), pwr_w_ray.size, inds_ray)


def get_ivic_pct(npulses, pct=1, prob_thr=1e-4):
    """
    Get the point clutter threshold (PCT)

    Parameters
    ----------
    npulses : 1D array
        array with the number of pulses in a ray
    pct : 1D array
        array with possible values of the PCT
    z : 1D array
        array of z values
    prob_thr : float
        Desired probability

    Returns
    -------
    pct_out : 1D array
        The PCT threshold corresponding to each number of pulses

    """
    x = np.arange(0.0001, 400., 0.01)
    pct_out = np.ma.masked_all(len(npulses))
    for i, npuls in enumerate(npulses):
        # default tolerance 1.49012e-8
        f = fsolve(
            _func_pct_int, pct, args=(x, npuls, prob_thr), xtol=1e-13)
        pct_out[i] = f

    return pct_out


def _func_pct_int(pct, x, npuls, prob_thr):
    """
    Function to solve for point clutter target

    Parameters
    ----------
    x : 1D array
        array with the variable in the integral
    pct : float
        The point clutter threshold
    npuls : float
        the number of pulses per ray

    Returns
    -------
    f : 1D array
        function f(x)

    """
    # default tolerance 1.49e-8
    f, _ = integrate.quad(
        _func_pct, 0., 400., args=(pct, npuls), epsabs=1e-10,
        epsrel=1e-06)

    return f-prob_thr


def _func_pct(x, pct, npuls):
    """
    Function to solve for point clutter target

    Parameters
    ----------
    x : 1D array
        array with the variable in the integral
    pct : float
        The point clutter threshold
    npuls : float
        the number of pulses per ray

    Returns
    -------
    f : 1D array
        function f(x)

    """
    log_e = np.log10(np.e)
    f = (
        2.*np.power(
            10.,
            np.log10(x)*(npuls-1)-log_e*(x+gammaln(npuls)) +
            np.log10(gammainc(npuls, x/pct))-log_e*(gammaln(npuls+1)+x/pct) +
            np.log10(x/pct)*npuls) -
        np.power(
            10.,
            np.log10(x)*(npuls-1)-log_e*(x+gammaln(npuls)) +
            2.*(np.log10(gammainc(npuls, x/pct)) -
                log_e*(gammaln(npuls+1)+(x/pct))+np.log10(x/pct)*npuls)))
    return f


def get_ivic_flat_reg_var_max(npulses, flat_reg_wlen, n=40., prob_thr=0.01):
    """
    Get the threshold for maximum local variance of noise [dB]

    Parameters
    ----------
    npulses : 1D array
        array with the number of pulses in a ray
    flat_reg_wlen : int
        the length of the flat region window in bins
    n : float
        the mean noise power
    prob_thr : float
        Probably of falsely detecting noise-only flat section as contaminated
        with signal

    Returns
    -------
    var_thr : 1D array
        The thresholds corresponding to each number of pulses

    """
    y = np.arange(0.0001, 1e4, 0.01)

    log_e = np.log10(np.e)
    log_log = np.log10(np.log(10.))

    wconst1 = flat_reg_wlen-1
    wconst2 = flat_reg_wlen-2+1/flat_reg_wlen
    wconst3 = flat_reg_wlen*flat_reg_wlen-3*flat_reg_wlen+5-3/flat_reg_wlen
    wconst4 = (
        (-2*flat_reg_wlen*flat_reg_wlen*flat_reg_wlen +
         12*flat_reg_wlen*flat_reg_wlen-22*flat_reg_wlen+12)/flat_reg_wlen)
    wconst5 = 4*(2-flat_reg_wlen-1/flat_reg_wlen)
    wconst6 = (
        (flat_reg_wlen-1)*(flat_reg_wlen-2)*(flat_reg_wlen-3)/flat_reg_wlen)

    var_thr = np.ma.masked_all(len(npulses))
    for i, npuls in enumerate(npulses):
        pdb_part1 = np.power(
            10., np.log10(npuls/n)*npuls + log_log-gammaln(npuls)*log_e)
        pdb = np.empty(4)
        for j in range(pdb.size):
            k = j+1

            # find first 0 crossing
            f_y = _func_flat_reg(y, k, npuls, n)
            ind = np.where(f_y != 0.)[0]
            cons_list = np.split(ind, np.where(np.diff(ind) != 1)[0]+1)
            ind = cons_list[0][-1]

            # solve integral
            pdb[j] = pdb_part1*integrate.quad(
                _func_flat_reg, 0., y[ind], args=(k, npuls, n))[0]

        pdb_1_2 = pdb[0]*pdb[0]

        vardb_1 = wconst1*(pdb[1]-pdb_1_2)
        vardb_2 = (
            pdb[3]*wconst2+pdb[1]*pdb[1]*wconst3+pdb[1]*pdb_1_2*wconst4 +
            pdb[0]*pdb[2]*wconst5+wconst6*pdb_1_2*pdb_1_2)

        vardb_1_2 = vardb_1*vardb_1

        alpha = vardb_1_2/(vardb_2-vardb_1_2)
        theta = (vardb_2-vardb_1_2)/vardb_1

        var_thr[i] = gammainccinv(alpha, prob_thr)*theta

    return var_thr


def _func_flat_reg(y, k, npuls, n):
    """
    Function to solve for flat region variance max. Derived from eq. B5 in
    paper

    Parameters
    ----------
    y : 1D array
        array with the variable in the integral
    k : int
        The order
    npuls : int
        the number of pulses per ray
    n : float
        noise power

    Returns
    -------
    f : 1D array
        the f(y) function

    """
    return np.power(
        10., k*np.log10(y)+y*npuls-np.power(10., y)*npuls/n*np.log10(np.e))


def get_ivic_snr_thr(npulses, pfa_thr=1e-3):
    """
    Get the threshold for steps 3 and 6 of ivic

    Parameters
    ----------
    npulses : 1D array
        array with the number of pulses in a ray
    pfa_thr : float
        the desired probability of false alarm

    Returns
    -------
    snr_thr : 1D array
        The snr threshold corresponding to each number of pulses according to
        the desired probability of false alarm


    """
    return gammainccinv(npulses, pfa_thr)/npulses


def ivic_pct_table(npulses):
    """
    Get the point clutter threshold (PCT) of Ivic from a look up table.
    The thresholds are computed for between 1 and 200 pulses. If there
    number of pulses is beyond this range it throws an error. The table is
    computed for a PFA = 1e-4.
    WARNING: This thresholds were computed by I. Ivic in Matlab. the values
    differ from those computed by get_ivic_pct, particularly for small values
    of npulses.

    Parameters
    ----------
    npulses : 1D array
        array with the number of pulses in a ray

    Returns
    -------
    pct_table : 1D array
        The PCT corresponding to each number of pulses

    """
    ind = np.where(np.logical_or(npulses > 200, npulses < 1))[0]
    if ind.size > 0:
        raise ValueError("Table valid for number of pulses between 1 and 200")

    ind_npulses = (npulses-1).astype(np.int)

    pct_table = np.array([
        0., 243.596101, 56.972570, 27.309237, 17.372310, 12.729620, 10.119393,
        8.470525, 7.342833, 6.526113, 5.908569, 5.425731, 5.038014, 4.719832,
        4.453971, 4.228437, 4.034628, 3.866220, 3.718462, 3.587719, 3.471161,
        3.366555, 3.272113, 3.186389, 3.108200, 3.036567, 2.970678, 2.909848,
        2.853498, 2.801135, 2.752336, 2.706737, 2.664023, 2.623917, 2.586180,
        2.550599, 2.516987, 2.485179, 2.455026, 2.426399, 2.399179, 2.373260,
        2.348546, 2.324953, 2.302401, 2.280820, 2.260146, 2.240321, 2.221289,
        2.203003, 2.185416, 2.168489, 2.152181, 2.136459, 2.121289, 2.106642,
        2.092489, 2.078804, 2.065563, 2.052745, 2.040327, 2.028291, 2.016617,
        2.005289, 1.994292, 1.983609, 1.973227, 1.963132, 1.953312, 1.943755,
        1.934451, 1.925388, 1.916557, 1.907949, 1.899554, 1.891365, 1.883373,
        1.875571, 1.867951, 1.860508, 1.853235, 1.846125, 1.839173, 1.832373,
        1.825720, 1.819209, 1.812835, 1.806593, 1.800480, 1.794491, 1.788622,
        1.782869, 1.777229, 1.771698, 1.766273, 1.760950, 1.755727, 1.750600,
        1.745567, 1.740624, 1.735771, 1.731003, 1.726318, 1.721715, 1.717190,
        1.712742, 1.708369, 1.704069, 1.699839, 1.695678, 1.691584, 1.687556,
        1.683591, 1.679689, 1.675847, 1.672064, 1.668338, 1.664669, 1.661055,
        1.657494, 1.653986, 1.650528, 1.647121, 1.643762, 1.640450, 1.637186,
        1.633966, 1.630792, 1.627660, 1.624572, 1.621524, 1.618518, 1.615551,
        1.612624, 1.609734, 1.606882, 1.604067, 1.601287, 1.598542, 1.595832,
        1.593156, 1.590512, 1.587901, 1.585322, 1.582774, 1.580256, 1.577769,
        1.575311, 1.572881, 1.570480, 1.568106, 1.565760, 1.563441, 1.561148,
        1.558880, 1.556638, 1.554421, 1.552228, 1.550059, 1.547913, 1.545791,
        1.543691, 1.541614, 1.539559, 1.537525, 1.535513, 1.533521, 1.531550,
        1.529599, 1.527668, 1.525757, 1.523864, 1.521991, 1.520136, 1.518299,
        1.516481, 1.514680, 1.512897, 1.511130, 1.509381, 1.507648, 1.505932,
        1.504232, 1.502548, 1.500879, 1.499226, 1.497588, 1.495965, 1.494356,
        1.492763, 1.491183, 1.489618, 1.488066, 1.486528, 1.485004, 1.483493,
        1.481995, 1.480511, 1.479038, 1.477579])

    return pct_table[ind_npulses]


def ivic_flat_reg_var_max_table(npulses):
    """
    Get maximum variance of noise of Ivic from a look up table.
    These values are computed for a prescribed flat region window length that
    depends on the number of pulses. The window length as a function of number
    of pulses can be found using function ivic_flat_reg_wind_len

    Parameters
    ----------
    npulses : 1D array
        array with the number of pulses in a ray
    flat_reg_wlen : float
        The flat region window length in bins

    Returns
    -------
    flat_reg_var_max : 1D array
        The maximum variance threshold corresponding to each number of pulses

    """
    ind = np.where(np.logical_or(npulses > 200, npulses < 1))[0]
    if ind.size > 0:
        raise ValueError("Table valid for number of pulses between 1 and 200")

    ind_npulses = (npulses-1).astype(np.int)

    flat_reg_var_max = np.array([
        0., 12.501880, 9.066873, 6.978578, 5.602964, 3.184561, 2.709687,
        2.352218, 2.075309, 1.855436, 1.677084, 1.529734, 1.406049, 1.300803,
        0.683347, 0.625970, 0.592077, 0.560521, 0.531522, 0.505351, 0.481602,
        0.459942, 0.439551, 0.420858, 0.403674, 0.387825, 0.373164, 0.359562,
        0.346910, 0.335113, 0.324087, 0.313760, 0.304067, 0.292595, 0.284680,
        0.277142, 0.269806, 0.262747, 0.256040, 0.249657, 0.243574, 0.237771,
        0.232228, 0.226885, 0.221765, 0.216871, 0.212188, 0.207704, 0.203405,
        0.199281, 0.195321, 0.191515, 0.187856, 0.184334, 0.180942, 0.177673,
        0.174515, 0.171464, 0.168519, 0.165675, 0.162927, 0.160270, 0.157700,
        0.155213, 0.152805, 0.150471, 0.148210, 0.146016, 0.143889, 0.141824,
        0.139819, 0.137871, 0.135979, 0.134139, 0.132349, 0.130609, 0.128915,
        0.127265, 0.125659, 0.124094, 0.122569, 0.121083, 0.119633, 0.118219,
        0.116839, 0.115493, 0.114178, 0.112894, 0.111640, 0.110415, 0.109218,
        0.108047, 0.106903, 0.105783, 0.104688, 0.103617, 0.102568, 0.101542,
        0.100537, 0.099552, 0.098588, 0.097644, 0.096718, 0.095811, 0.094921,
        0.094049, 0.093194, 0.092355, 0.091059, 0.090263, 0.089482, 0.088715,
        0.087962, 0.087223, 0.086497, 0.085783, 0.085083, 0.084394, 0.083718,
        0.083053, 0.082399, 0.081757, 0.081125, 0.080504, 0.079893, 0.079292,
        0.078700, 0.078119, 0.077546, 0.076983, 0.076428, 0.075882, 0.075345,
        0.074816, 0.074295, 0.073781, 0.073276, 0.072778, 0.072287, 0.071804,
        0.071327, 0.070858, 0.070395, 0.069939, 0.069489, 0.069046, 0.068609,
        0.068178, 0.067753, 0.067334, 0.066920, 0.066512, 0.066110, 0.065713,
        0.065321, 0.064934, 0.064553, 0.064176, 0.063804, 0.063437, 0.063075,
        0.062717, 0.062364, 0.062015, 0.061671, 0.061331, 0.060995, 0.060663,
        0.060335, 0.060011, 0.059691, 0.059375, 0.059063, 0.058754, 0.058449,
        0.058148, 0.057850, 0.057555, 0.057262, 0.056972, 0.056686, 0.056403,
        0.056123, 0.055846, 0.055573, 0.055302, 0.055034, 0.054770, 0.054508,
        0.054249, 0.053993, 0.053740, 0.053489, 0.053241, 0.052996, 0.052753,
        0.052513, 0.052276, 0.052040, 0.051808])

    return flat_reg_var_max[ind_npulses]


def ivic_flat_reg_wind_len_table(npulses):
    """
    Get the size of the flat region window length as a function of the number
    of pulses.

    Parameters
    ----------
    npulses : 1D array
        array with the number of pulses in a ray

    Returns
    -------
    flat_reg_var_max : 1D array
        The maximum variance threshold corresponding to each number of pulses

    """
    ind = np.where(np.logical_or(npulses > 200, npulses < 1))[0]
    if ind.size > 0:
        raise ValueError("Table valid for number of pulses between 1 and 200")

    ind_npulses = (npulses-1).astype(np.int)

    flat_reg_wlen = np.zeros(200, dtype=np.int)
    flat_reg_wlen[0:5] += 100
    flat_reg_wlen[5:14] += 64
    flat_reg_wlen[14:] += 32

    return flat_reg_wlen[ind_npulses]


def ivic_snr_thr_table(npulses):
    """
    Get the threshold for steps 3 and 6 of ivic from a look up table
    The thresholds are computed for between 1 and 200 pulses. If there
    number of pulses is beyond this range it throws an error. The table is
    computed for a PFA = 1e-3.

    Parameters
    ----------
    npulses : 1D array
        array with the number of pulses in a ray

    Returns
    -------
    snr_thr_out : 1D array
        The snr threshold corresponding to each number of pulses according to
        the desired probability of false alarm


    """
    ind = np.where(np.logical_or(npulses > 200, npulses < 1))[0]
    if ind.size > 0:
        raise ValueError("Table valid for number of pulses between 1 and 200")

    ind_npulses = (npulses-1).astype(np.int)

    snr_thr_table = np.array([
        0., 4.616707, 3.742957, 3.265560, 2.958830, 2.742458, 2.580234,
        2.453272, 2.350689, 2.265737, 2.193997, 2.132442, 2.078922,
        2.031867, 1.990102, 1.952726, 1.919036, 1.888477, 1.860602,
        1.835049, 1.811518, 1.789762, 1.769572, 1.750774, 1.733216,
        1.716772, 1.701330, 1.686795, 1.673083, 1.660121, 1.647843,
        1.636193, 1.625119, 1.614578, 1.604528, 1.594932, 1.585759,
        1.576978, 1.568564, 1.560490, 1.552737, 1.545282, 1.538108,
        1.531199, 1.524537, 1.518110, 1.511904, 1.505906, 1.500106,
        1.494493, 1.489057, 1.483789, 1.478682, 1.473726, 1.468916,
        1.464243, 1.459703, 1.455287, 1.450992, 1.446812, 1.442742,
        1.438776, 1.434912, 1.431144, 1.427469, 1.423883, 1.420383,
        1.416964, 1.413625, 1.410363, 1.407173, 1.404054, 1.401004,
        1.398019, 1.395097, 1.392237, 1.389436, 1.386692, 1.384004,
        1.381369, 1.378785, 1.376252, 1.373767, 1.371330, 1.368937,
        1.366589, 1.364284, 1.362021, 1.359798, 1.357614, 1.355468,
        1.353359, 1.351286, 1.349248, 1.347244, 1.345273, 1.343334,
        1.341427, 1.339550, 1.337703, 1.335884, 1.334094, 1.332332,
        1.330596, 1.328886, 1.327202, 1.325543, 1.323908, 1.322296,
        1.320708, 1.319142, 1.317599, 1.316077, 1.314576, 1.313096,
        1.311635, 1.310195, 1.308774, 1.307371, 1.305987, 1.304621,
        1.303273, 1.301942, 1.300627, 1.299330, 1.298048, 1.296783,
        1.295533, 1.294298, 1.293078, 1.291873, 1.290682, 1.289506,
        1.288343, 1.287193, 1.286057, 1.284934, 1.283824, 1.282726,
        1.281641, 1.280568, 1.279506, 1.278456, 1.277418, 1.276391,
        1.275375, 1.274370, 1.273375, 1.272391, 1.271417, 1.270454,
        1.269500, 1.268556, 1.267622, 1.266697, 1.265781, 1.264875,
        1.263978, 1.263089, 1.262209, 1.261338, 1.260475, 1.259620,
        1.258774, 1.257935, 1.257105, 1.256282, 1.255467, 1.254659,
        1.253859, 1.253066, 1.252281, 1.251502, 1.250731, 1.249966,
        1.249209, 1.248458, 1.247713, 1.246975, 1.246244, 1.245518,
        1.244799, 1.244086, 1.243380, 1.242679, 1.241984, 1.241295,
        1.240611, 1.239933, 1.239261, 1.238594, 1.237933, 1.237277,
        1.236626, 1.235981, 1.235341, 1.234705, 1.234075, 1.233450,
        1.232829])

    return snr_thr_table[ind_npulses]



"""
pyart.aux_io.mf_dat_reader
==========================

Routines for putting MeteoFrance operational radar data contained in text
files into grid object.

.. autosummary::
    :toctree: generated/

    read_dat_mf

"""

import os
import datetime
from warnings import warn

import numpy as np

from ..config import FileMetadata
from ..io.common import _test_arguments
from ..core.grid import Grid
from ..util import ma_broadcast_to

DAT_FIELD_NAMES = {
    'RR': 'radar_estimated_rain_rate'
}


def read_dat_mf(filename, additional_metadata=None, xres=1., yres=1., nx=1536,
                ny=1536, nz=1, nd=65535,
                field_name='radar_estimated_rain_rate', **kwargs):

    """
    Read a MeteoFrance operational radar data binary file. The data is in
    stereopolar projection

    Parameters
    ----------
    filename : str
        Name of the file to read.
    additional_metadata : dict of dicts, optional
        Dictionary of dictionaries to retrieve metadata during this read.
        This metadata is not used during any successive file reads unless
        explicitly included.  A value of None, the default, will not
        introduct any addition metadata and the file specific or default
        metadata as specified by the Py-ART configuration file will be used.
    xres, yres : float
        resolution of each grid point [km]
    nx, ny, nz : int
        dimensions of the grid
    nd : int
        integer representing No Data
    field_name : str
        name of the field stored in the binary file

    Returns
    -------
    grid : Grid
        Grid object containing data the data.

    """
    # test for non empty kwargs
    _test_arguments(kwargs)

    try:
        with open(filename, 'r') as file:
            data = np.loadtxt(file, comments='#')
            data = np.ma.masked_equal(data, nd)
            # conversion to mm/h
            # original data in 0.1 mm/5 min
            data = 1.2*data
            data = data[::-1, :]
    except EnvironmentError as ee:
        warn(str(ee))
        warn('Unable to read file '+filename)
        return None, None

    # reserved_variables = [
    #     'time', 'x', 'y', 'z',
    #     'origin_latitude', 'origin_longitude', 'origin_altitude',
    #     'point_x', 'point_y', 'point_z', 'projection',
    #     'point_latitude', 'point_longitude', 'point_altitude',
    #     'radar_latitude', 'radar_longitude', 'radar_altitude',
    #     'radar_name', 'radar_time', 'base_time', 'time_offset',
    #     'ProjectionCoordinateSystem']

    # metadata
    metadata = dict()

    filemetadata = FileMetadata('DAT', DAT_FIELD_NAMES, additional_metadata)

    # required reserved variables
    time = filemetadata('grid_time')
    origin_latitude = filemetadata('origin_latitude')
    origin_longitude = filemetadata('origin_longitude')
    origin_altitude = filemetadata('origin_altitude')
    x = filemetadata('x')
    y = filemetadata('y')
    z = filemetadata('z')

    # Nortwest corner
    # -619652.074056 -3526818.337932 m
    # -9.965 53.670 NW (deg)

    x_vals = 1000.*(np.arange(nx)*xres+xres/2.)-619652.074056
    y_vals = -3526818.337932-1000.*(np.arange(ny)*yres+yres/2.)
    x['data'] = x_vals
    y['data'] = y_vals[::-1]
    z['data'] = np.array([0.])

    # origin of stereo-polar projection
    origin_latitude['data'] = np.array([90.])
    origin_longitude['data'] = np.array([0.])
    origin_altitude['data'] = np.array([0.])

    # projection (stereo-polar)
    projection = {
        'proj': 'stere',
        'lat_ts': 45.,
        'ellps': 'WGS84',
        'datum': 'WGS84',
        'lat_0': 90.,
        'lon_0': 0.
    }

    bfile = os.path.basename(filename)
    # Time
    prod_time = datetime.datetime.strptime(bfile[:12], '%Y%m%d%H%M')
    time['units'] = 'seconds since '+prod_time.strftime('%Y-%m-%d %H:%M:%S')
    time['data'] = np.array([0.])

    # read in the fields
    fields = {}
    field_dict = filemetadata(field_name)
    field_dict['data'] = ma_broadcast_to(data, (nz, ny, nx))
    fields[field_name] = field_dict

    # radar variables
    radar_latitude = None
    radar_longitude = None
    radar_altitude = None
    radar_name = None
    radar_time = None

    return Grid(
        time, fields, metadata,
        origin_latitude, origin_longitude, origin_altitude, x, y, z,
        projection=projection,
        radar_latitude=radar_latitude, radar_longitude=radar_longitude,
        radar_altitude=radar_altitude, radar_name=radar_name,
        radar_time=radar_time)



"""
pyart.config
============

Py-ART configuration.

.. autosummary::
    :toctree: generated/

    load_config
    get_metadata
    get_fillvalue
    get_field_name
    get_field_colormap
    get_field_limits
    get_field_mapping
    FileMetadata

"""

import os
import traceback
import warnings
import fnmatch
import numpy as np

# the path to the default configuration file
_dirname = os.path.dirname(__file__)
_DEFAULT_CONFIG_FILE = os.path.join(_dirname, 'default_config.py')


def load_config(filename=None):
    """
    Load a Py-ART configuration from a config file.

    The default values for a number of Py-ART parameters and metadata is
    controlled by a single Python configuration file. An self-descriping
    example of this file can be found in the Py-ART source directory named
    **default_config.py**. These defaults can modified by setting the
    environmental variable `PYART_CONFIG` to point to a new configuration
    file. If this variable is not set then the settings contained in
    the **default_config.py** file are used.

    The code the configuration file is executed as-is with full permission,
    this may present a security issue, do not load un-trusted configuration
    files.

    The recommended method for changing these defaults is for users to
    copy this file into their home directory, rename it to .pyart_config.py,
    make any changes, and adjust their login scripts to set the PYART_CONFIG
    environmental variable to point to .pyart_config.py in their home
    directory.

    Py-ART's configuration can also be modified within a script or shell
    session using this function, the modification will last until a the end
    of the script/session or until a new configuration is loaded.

    Parameters
    ----------
    filename : str
        Filename of configuration file. If None the default configuration
        file is loaded from the Py-ART source code directory.

    """
    if filename is None:
        filename = _DEFAULT_CONFIG_FILE

    # these are private since they should not be accessed by users or other
    # modules, use the get_ functions.
    global _DEFAULT_METADATA
    global _FILE_SPECIFIC_METADATA
    global _FIELD_MAPPINGS
    global _FILL_VALUE
    global _DEFAULT_FIELD_NAMES
    global _DEFAULT_FIELD_COLORMAP
    global _DEFAULT_FIELD_LIMITS

    try:
        from importlib.machinery import SourceFileLoader
        cfile = SourceFileLoader('metadata_config', filename).load_module()
    except ImportError:
        import imp
        cfile = imp.load_source('metadata_config', filename)

    _DEFAULT_METADATA = cfile.DEFAULT_METADATA
    _FILE_SPECIFIC_METADATA = cfile.FILE_SPECIFIC_METADATA
    _FIELD_MAPPINGS = cfile.FIELD_MAPPINGS
    _FILL_VALUE = cfile.FILL_VALUE
    _DEFAULT_FIELD_NAMES = cfile.DEFAULT_FIELD_NAMES
    _DEFAULT_FIELD_COLORMAP = cfile.DEFAULT_FIELD_COLORMAP
    _DEFAULT_FIELD_LIMITS = cfile.DEFAULT_FIELD_LIMITS
    return

# load the configuration from the enviromental parameter if it is set
# if the load fails issue a warning and load the default config.
_config_file = os.environ.get('PYART_CONFIG')
if _config_file is None:
    load_config(_DEFAULT_CONFIG_FILE)
else:
    try:
        load_config(_config_file)
    except:
        msg = ("\nLoading configuration from PYART_CONFIG enviromental "
               "variable failed:"
               "\n--- START IGNORED TRACEBACK --- \n" +
               traceback.format_exc() +
               "\n --- END IGNORED TRACEBACK ---"
               "\nLoading default configuration")
        warnings.warn(msg)
        load_config(_DEFAULT_CONFIG_FILE)


def get_metadata(p):
    """
    Return a dictionary of metadata for a given parameter, p.

    An empty dictionary will be returned in no metadata dictionary exists for
    parameter p.
    """
    if p in _DEFAULT_METADATA:
        return _DEFAULT_METADATA[p].copy()
    else:
        return {}


def get_fillvalue():
    """
    Return the current fill value.
    """
    return _FILL_VALUE


def get_field_name(field):
    """
    Return the field name from the configuration file for a given field.
    """
    return str(_DEFAULT_FIELD_NAMES[field])


def get_field_colormap(field):
    """
    Return the colormap name from the configuration file for a field name.
    """
    if field in _DEFAULT_FIELD_COLORMAP:
        return _DEFAULT_FIELD_COLORMAP[field]
    else:
        import matplotlib.cm
        return matplotlib.cm.get_cmap().name


def get_field_limits(field, container=None, selection=0):
    """
    Return the data limits from the configuration file for a given field,
    radar and sweep.

    Parameters
    ----------
    field : str
        Field name.
    container : Radar, Grid or None, optional
        This is an optional parameter that will be use to get informations
        related to the field, like for instace nyquist velocity.
    selection : int, optional
        Selection of the data in the container, case container is a Radar this
        is the sweep to be considered.

    Returns
    -------
    vmin, vmax: 2-tuplet of float
        Minimun and Maximun teorical value for field, if field is not
        in the configuration file returns (None, None).

    """
    if field in _DEFAULT_FIELD_LIMITS:
        limits = _DEFAULT_FIELD_LIMITS[field]
        if callable(limits):
            limits = limits(container, selection)
        return limits
    else:
        return None, None


def get_field_mapping(filetype):
    """
    Return a copy of the default field mapping for a given file type.

    Parameters
    ----------
    filetype : str
        Filetype to return field mappings for.

    Returns
    -------
    field_mappings : dict
        Dictionary mapping field names from one type to another.

    """
    return _FIELD_MAPPINGS[filetype].copy()


class FileMetadata():
    """
    A class for accessing metadata needed when reading files.

    Parameters
    ----------
    filetype: str
        Type of file being read.
    field_names : dict
        Dictionary mapping file field names to radar fields names.
    additional_metadata : dict of dicts
        Additional metadata to use during read.
    file_field_names : bool
        True to keep the field names in the file.
    exclude_fields : list of strings
        Fields to exclude during readings.
    include_fields : list of strings
        Fields to include during readings.

    """

    def __init__(self, filetype, field_names=None, additional_metadata=None,
                 file_field_names=False, exclude_fields=None,
                 include_fields=None):
        """
        Initialize.
        """
        # parse filetype parameter
        if filetype in _FILE_SPECIFIC_METADATA:
            self._file_specific_metadata = _FILE_SPECIFIC_METADATA[filetype]
        else:
            self._file_specific_metadata = {}

        # parse additional_metadata
        if additional_metadata is None:
            self._additional_metadata = {}
        else:
            self._additional_metadata = additional_metadata

        # parse field_names and file_field_names
        if file_field_names:
            self._field_names = None
        elif field_names is None and filetype in _FIELD_MAPPINGS:
            # if filetype is missing there is no mapping
            self._field_names = _FIELD_MAPPINGS[filetype]
        else:
            self._field_names = field_names

        # parse exclude_fields
        if exclude_fields is None:
            self._exclude_fields = []
        else:
            self._exclude_fields = exclude_fields

        if include_fields is None:
            self._include_fields = None
        else:
            self._include_fields = include_fields

    def get_metadata(self, p):
        """
        Retrieve metadata for a parameter `p`.

        Parameters
        ----------
        p : str
            Parameter to retrieve metadata for.

        Returns
        -------
        dic : dict
            Dictionary of metadata for the parameter.

        """
        # additional_metadata is queued first
        if p in self._additional_metadata:
            return self._additional_metadata[p].copy()

        # then the file specific metadata
        elif p in self._file_specific_metadata:
            return self._file_specific_metadata[p].copy()

        # and finally the default metadata
        elif p in _DEFAULT_METADATA:
            return _DEFAULT_METADATA[p].copy()

        # return a empty dict if the parameter is in none of the above
        else:
            return {}

    def __call__(self, p):
        """
        Retrieve metadata for parameter `p`.
        """
        return self.get_metadata(p)

    def get_field_name(self, file_field_name):
        """
        Return the name radar field for a given file field name

        Parameters
        ----------
        file_field_name : str
            Field name in file being read.

        Returns
        -------
        field_name : str or None
            Field name in radar object fields dictionary, None indicated
            that the field should not be included.

        """
        
        if self._field_names is None:
            field_name = file_field_name
        elif any([fnmatch.fnmatch(file_field_name,fn) 
                  for fn in self._field_names]):
            idx = np.where([fnmatch.fnmatch(file_field_name,fn)
                  for fn in self._field_names])[0][0]
            field_name = [x for x in self._field_names.values()][idx]
        else:
            return None # field is not mapped
        
        if field_name in self._exclude_fields:
            return None # field is excluded
        elif self._include_fields is not None:
            if not field_name in self._include_fields:
                return None
            else:
                return field_name
        else:
            return field_name
